/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.latentVariables;

import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Locale;
import java.util.Random;

import javax.naming.OperationNotSupportedException;

import de.jstacs.NotTrainedException;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.EmptyDataSetException;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.io.NonParsableException;
import de.jstacs.io.XMLParser;
import de.jstacs.projects.inmode.models.variableStructure.AbstractVariableStructureModel;
import de.jstacs.projects.inmode.utils.BasicSamplingAlgorithms;
import de.jstacs.projects.inmode.utils.NML;
import de.jstacs.results.NumericalResultSet;
import de.jstacs.sequenceScores.statisticalModels.trainable.AbstractTrainableStatisticalModel;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.utils.Normalisation;

/**
 * An OOPS (one occurrence per sequence model) that uses a mixture of {@link AbstractVariableStructureModel}s as motif components.
 * Contains traditional mixture model and standard OOPS model as special cases.
 * @author Ralf Eggeling
 * 
 */
public class MultiComponentOOPSModel extends AbstractLatentVariableModel {
	

	private static final String XML_TAG = "MultiComponentOOPSModel";

	private AbstractVariableStructureModel[] motifModel;
	
	private AbstractTrainableStatisticalModel flankingModel;
		
	private double[] strandOrientationLogProbabilityParameter;
	
	private double[] motifTypeLogProbabilityParameter;
	
	private boolean[] activeComponents;
	
	protected int numberOfInitialIterations;
	
	protected int numberOfAdditionalIterations;
	
	protected int numberOfRestarts;
	
	
	//static variables used during optimization
	private static DataSet[] currentBindingSites;
	private static double[][] currentWeights;
	private static int[] bestLV;
	private static double maxScore;
	private static int bestRs;
	private static int bestIt;
	private static int[] bestLVLocal;
	private static int lastOptimalModelUpdate;
	private static int it;
	private static double maxScoreLocal;
	private static double currentProgress;
	private static boolean fullyObservable=false;
	private static boolean dataAligned=false;
	private static boolean updateLVParams=false;

	private ProgressUpdater progress;
	private Protocol protocol;
	private StringBuffer log;
	private NumberFormat nf;
	
	private static boolean textOutput = true;
	
	private double jointStructureScore;
	
	
	/**
	 * This is the main constructor.
	 * @param motifModel An arbitrary number of {@link AbstractVariableStructureModel}s that are used for representing the motif.
	 * @param flankingModel An model for the flanking (non-motif) sequences. 
	 * @param bothStrands Indicates whether the motif can be located on both strands or not.
	 * @param numberOfInitialIterations The number of iteration steps that is always to be performed per restarts.
	 * @param numberOfAdditionalIterations The number of iteration steps that need to pass since last improvement of score before termination.
	 * @param numberOfRestarts The number of independent restarts that need to be run.
	 */
	public MultiComponentOOPSModel(AbstractVariableStructureModel[] motifModel, AbstractTrainableStatisticalModel flankingModel, boolean bothStrands, int numberOfInitialIterations, int numberOfAdditionalIterations, int numberOfRestarts) {
		super(new AlphabetContainer(DNAAlphabet.SINGLETON),0);
		this.numberOfInitialIterations=numberOfInitialIterations;
		this.numberOfAdditionalIterations=numberOfAdditionalIterations;
		this.numberOfRestarts=numberOfRestarts;
		this.motifModel = motifModel;
		this.flankingModel = flankingModel;
		if(bothStrands) {
			strandOrientationLogProbabilityParameter  = new double[]{Math.log(0.5),Math.log(0.5)};
		} else {
			strandOrientationLogProbabilityParameter = new double[]{0};
		}
		motifTypeLogProbabilityParameter = new double[motifModel.length];
		activeComponents = new boolean[motifModel.length];
		Arrays.fill(activeComponents, true);
		Arrays.fill(motifTypeLogProbabilityParameter,Math.log(1d/motifModel.length));
		currentWeights = new double[motifModel.length][];
		currentBindingSites = new DataSet[motifModel.length];
		this.length=0;
		textOutput=false;
		updateLVParams=false;
		nf = NumberFormat.getInstance(Locale.US);
		nf.setGroupingUsed(false);
		nf.setMaximumFractionDigits(2);
		nf.setMinimumFractionDigits(2);
	}
	
	/**
	 * The standard constructor for the interface {@link de.jstacs.Storable}. Creates a new {@link MultiComponentOOPSModel} out of a {@link StringBuffer}.
	 * @param sb the {@link StringBuffer} to be parsed
	 * @throws NonParsableException  if the {@link MultiComponentOOPSModel} could not be reconstructed out of
	 *             the {@link StringBuffer} <code>sb</code>      
	 */
	public MultiComponentOOPSModel(StringBuffer sb) throws NonParsableException {
		super(XMLParser.extractForTag(sb,XML_TAG));
	}
	
	
	protected void fromXML(StringBuffer xml) throws NonParsableException {
		this.alphabets = XMLParser.extractObjectForTags(xml, "alphabets", AlphabetContainer.class);
		this.length = XMLParser.extractObjectForTags(xml, "length", int.class);
		this.flankingModel =  XMLParser.extractObjectForTags(xml, "flankingModel", AbstractTrainableStatisticalModel.class);
		this.motifModel = XMLParser.extractObjectForTags(xml, "motifModel", AbstractVariableStructureModel[].class);
		this.activeComponents = XMLParser.extractObjectForTags(xml, "activeComponents", boolean[].class);
		this.strandOrientationLogProbabilityParameter = XMLParser.extractObjectForTags(xml, "strandParameter", double[].class);
		this.motifTypeLogProbabilityParameter = XMLParser.extractObjectForTags(xml, "typeParameter", double[].class);
		this.numberOfInitialIterations =  XMLParser.extractObjectForTags(xml, "numberOfInitialIterations", int.class);
		this.numberOfAdditionalIterations =  XMLParser.extractObjectForTags(xml, "numberOfAdditionalIterations", int.class);
		this.numberOfRestarts = XMLParser.extractObjectForTags(xml, "numberOfRestarts", int.class);
		this.jointStructureScore = XMLParser.extractObjectForTags(xml, "jointStructureScore",double.class);
	}
	
	
	public StringBuffer toXML() {
		StringBuffer xml = new StringBuffer();
		XMLParser.appendObjectWithTags(xml, alphabets, "alphabets");
		XMLParser.appendObjectWithTags(xml, length, "length");
		XMLParser.appendObjectWithTags(xml, flankingModel, "flankingModel");
		XMLParser.appendObjectWithTags(xml, motifModel, "motifModel");
		XMLParser.appendObjectWithTags(xml, activeComponents, "activeComponents");
		XMLParser.appendObjectWithTags(xml, strandOrientationLogProbabilityParameter, "strandParameter");
		XMLParser.appendObjectWithTags(xml, motifTypeLogProbabilityParameter, "typeParameter");
		XMLParser.appendObjectWithTags(xml, numberOfInitialIterations, "numberOfInitialIterations");
		XMLParser.appendObjectWithTags(xml, numberOfAdditionalIterations, "numberOfAdditionalIterations");
		XMLParser.appendObjectWithTags(xml, numberOfRestarts, "numberOfRestarts");
		XMLParser.appendObjectWithTags(xml, jointStructureScore, "jointStructureScore");
		XMLParser.addTags(xml, XML_TAG);
		return xml;
	}
	
	public MultiComponentOOPSModel clone() throws CloneNotSupportedException {
		MultiComponentOOPSModel m = (MultiComponentOOPSModel) super.clone();
		m.activeComponents = activeComponents.clone();
		m.flankingModel=flankingModel.clone();
		m.motifModel = new AbstractVariableStructureModel[motifModel.length];
		for(int c=0;c<motifModel.length;c++) 
			m.motifModel[c]=motifModel[c].clone();
		m.strandOrientationLogProbabilityParameter=strandOrientationLogProbabilityParameter.clone();
		m.motifTypeLogProbabilityParameter=motifTypeLogProbabilityParameter.clone();
		return m;
	}
	
		
	public void train(DataSet data, double[] weights) throws Exception {
		log = new StringBuffer();
		currentProgress = 0;
		
		if(this.motifModel.length==1&&data.getElementLength()==motifModel[0].getLength()&&this.strandOrientationLogProbabilityParameter.length==1) {
			numberOfInitialIterations=1;
			numberOfRestarts = 1;
			numberOfAdditionalIterations=0;
			protocol.append("Data is fully observable. Proceed with exact learning...");
			fullyObservable=true;
		} else {
			fullyObservable=false;
		}
		if(data.getElementLength()==motifModel[0].getLength()) {
			dataAligned=true;
		} else {
			dataAligned=false;
		}
			
		
		if(weights == null) {
			weights = new double[data.getNumberOfElements()];
			Arrays.fill(weights,1);
		}
				
		int[] lv = new int[data.getNumberOfElements()];
		Random rand = new Random();
		if(!flankingModel.isInitialized())
			flankingModel.train(data,weights);
		
		if(data.getMinimalElementLength()<motifModel[0].getLength())
			throw new Exception("Each input sequence needs to be at least as long as the width of the motif.");	
		
		maxScore = Double.NEGATIVE_INFINITY;
		bestRs = 0;
		bestIt = 0;
		
		for(int i=0;i<data.getNumberOfElements();i++)
			data.getElementAt(i).reverseComplement();
	
		for(int rs = 0;rs<numberOfRestarts;rs++) {
			//initialization
			
			for(int i=0;i<data.getNumberOfElements();i++) {
				int sL = data.getElementAt(i).getLength()-motifModel[0].getLength()+1;
				int position = BasicSamplingAlgorithms.sampleDiscreteRandomVariableFromUniformDistribution(sL, rand);
				int type = BasicSamplingAlgorithms.sampleDiscreteRandomVariableFromLogProbs(motifTypeLogProbabilityParameter, rand);
				int strand= BasicSamplingAlgorithms.sampleDiscreteRandomVariableFromLogProbs(strandOrientationLogProbabilityParameter, rand);
				lv[i] =getLV(position, type, strand);
			}		
			maxScoreLocal = Double.NEGATIVE_INFINITY;
			if(textOutput&&(!fullyObservable)) {
				protocol.append("\nRestart "+rs+":");
				log.append("Restart "+rs+"\n");
			}			
			double maxScoreLocal = getLVAssignment(data, weights, lv);
			if(maxScoreLocal>maxScore) {
				maxScore=maxScoreLocal;
				bestIt = lastOptimalModelUpdate;
				bestRs = rs;
				bestLV = bestLVLocal.clone();
			}
			
			
		}
		jointStructureScore = maxScore;
		this.updateDataForLatentVariables(data, weights, bestLV);
		for(int k=0;k<motifModel.length;k++) {
			if(activeComponents[k])
				motifModel[k].train(currentBindingSites[k],weights);
		}
		if(updateLVParams)
			updateLVParameters(bestLV,weights);
		if(textOutput&&(!fullyObservable)) {
			protocol.append("\nLearning completed:\n");
			protocol.append("Best model seen at restart "+bestRs+", iteration step "+bestIt+", with score "+nf.format(jointStructureScore)+"\n");
			if(motifModel.length==1) {
				protocol.append("Learned model:\n");
				protocol.append(motifModel[0].getInstanceName()+" of length "+motifModel[0].getLength()+" with "+motifModel[0].getNumericalCharacteristics().getResultAt(0).getValue()+" parameters\n");
			} else{
				protocol.append("Learned models:\n");
				NumberFormat nf = NumberFormat.getNumberInstance(Locale.US);
				nf.setMaximumFractionDigits(4);
				for(int k=0;k<motifModel.length;k++) {
					if(activeComponents[k])
						protocol.append((k+1)+"-th model: "+ motifModel[k].getInstanceName()+" of length "+motifModel[k].getLength()+" with "+motifModel[k].getNumericalCharacteristics().getResultAt(0).getValue()+" parameters and occurence probability "+nf.format(Math.exp(motifTypeLogProbabilityParameter[k]))+"\n");
					else
						protocol.append((k+1)+"-th model shrinked.\n");
				}
					
			}			
		}
	}
	
	private double getLVAssignment(DataSet data, double[] weights, int[] lv) throws Exception {
		Random rand = new Random();
		double previousScore = 0;
		lastOptimalModelUpdate = 0;
		if(!fullyObservable)
			protocol.append(" Initial phase...");
		boolean add = false;
		if(textOutput&&(!fullyObservable)) {
			log.append("It.step\tScore \tDifference\tlastOptimalStep\n");
		}
		for(it = 0; it<numberOfInitialIterations || ((it-lastOptimalModelUpdate)<numberOfAdditionalIterations);it++) {
			if(textOutput&&(!fullyObservable))
				log.append(it+"\t");
			if(it==numberOfInitialIterations) {
				if(!fullyObservable)
				protocol.append(" Additional phase...");
				add=true;
			}
			//get update data according to LV asssignment
			this.updateDataForLatentVariables(data, weights, lv);

			//learn motif models
			for(int k=0;k<motifModel.length;k++) {
				if(activeComponents[k])
					motifModel[k].train(currentBindingSites[k],weights);
			}
			/*	this.estimateHVParameters(lv);		*/
			if(updateLVParams)
				updateLVParameters(lv,weights);
			
			double score = 0;
			for(int k=0;k<motifModel.length;k++) {
				if(activeComponents[k])
					score+=2*motifModel[k].getLogScoreOfCurrentStructure();
			}
			
			// test for improvement to solution for current restart
			if(score>maxScoreLocal) {
				maxScoreLocal=score;
				lastOptimalModelUpdate = it;
				bestLVLocal = lv.clone();
			}
			if(textOutput&&(!fullyObservable)) {
				log.append(nf.format(score)+"\t"+nf.format((score-previousScore))+"\t"+lastOptimalModelUpdate);
			}
			
			// sample new LV according to current model
			for(int i=0;i<data.getNumberOfElements();i++) {
				double[] profile = getUnnormalizedSequenceProfile(data.getElementAt(i));
				Normalisation.logSumNormalisation(profile);
				lv[i] = BasicSamplingAlgorithms.sampleDiscreteRandomVariable(profile, rand);
			}			

			previousScore=score;
			if(textOutput&&(!fullyObservable))
				log.append("\n");
			
			if(it<numberOfInitialIterations) {
				currentProgress++;
				progress.setCurrent(currentProgress);
			} 
			
			if(it==numberOfInitialIterations-1) {
				protocol.append(" completed.");
			}
		}
		if(add==false) {
			if(!fullyObservable)
			protocol.append(" No additional phase required.");
		} else {
			protocol.append(" completed.");
		}
		return maxScoreLocal;
	}
	
	
	private void updateDataForLatentVariables(DataSet data, double[] weights, int[] lv) throws EmptyDataSetException, WrongAlphabetException {
		Sequence[][] newBS = new Sequence[motifModel.length][];
		for(int k=0;k<motifModel.length;k++) {
			int numberOfBS=0;
			for(int i=0;i<lv.length;i++) {
				if(getType(lv[i])==k)
					numberOfBS++;
			}
			newBS[k] = new Sequence[numberOfBS];
			currentWeights[k] = new double[numberOfBS];
		}
		int[] n = new int[motifModel.length];
		Arrays.fill(n, 0);
		for(int i=0;i<data.getNumberOfElements();i++) {
			int type = getType(lv[i]);
			int position = getPosition(lv[i]);
			int strand = getStrand(lv[i]);
			Sequence seq = data.getElementAt(i).getSubSequence(position, motifModel[type].getLength());
			if(strand==0) {
				newBS[type][n[type]] = seq;
			} else {
				try {
					newBS[type][n[type]] = seq.reverseComplement();
				} catch(OperationNotSupportedException e) {
					protocol.append("Warning: Reverse complement not supported. Use forward strand instead.\n");
					newBS[type][n[type]] = seq;
				}
			}
			currentWeights[type][n[type]]=weights[i];
			n[type]++;
		}
		
		for(int k=0;k<motifModel.length;k++) {
			if(n[k]==0) {
				activeComponents[k]=false;
				this.motifTypeLogProbabilityParameter[k]=Double.NEGATIVE_INFINITY;
			} else {
				activeComponents[k]=true;
				currentBindingSites[k] = new DataSet("",newBS[k]);
			}
		}
		Normalisation.logSumNormalisation(motifTypeLogProbabilityParameter);
		for(int k=0;k<motifModel.length;k++) {
			motifTypeLogProbabilityParameter[k] = Math.log(motifTypeLogProbabilityParameter[k]);
		}
	}
	
	private void updateLVParameters(int[] lv,double[] weights) {
		double forward = 0;
		double backward = 0;

		double[] mt = new double[motifModel.length];
		Arrays.fill(mt,0);
		for(int i=0;i<lv.length;i++) {
			mt[getType(lv[i])]+=weights[i];
			if(strandOrientationLogProbabilityParameter.length==2) {
				if(getStrand(lv[i]) == 0) {
					forward+=weights[i];
					} else {
					backward+=weights[i];
				}
			}
		} 
		if(strandOrientationLogProbabilityParameter.length==2) {
			strandOrientationLogProbabilityParameter = new double[]{NML.fsNMLeTerm(forward)*(forward+1),NML.fsNMLeTerm(backward)*(backward+1)};
			Normalisation.sumNormalisation(strandOrientationLogProbabilityParameter);
			strandOrientationLogProbabilityParameter[0] = Math.log(strandOrientationLogProbabilityParameter[0]);
			strandOrientationLogProbabilityParameter[1] = Math.log(strandOrientationLogProbabilityParameter[1]);
		}

		for(int k=0;k<motifModel.length;k++) {
			if(activeComponents[k])
				motifTypeLogProbabilityParameter[k] = NML.fsNMLeTerm(mt[k])*(mt[k]+1);
			else
				motifTypeLogProbabilityParameter[k]=0;
		}
		Normalisation.sumNormalisation(motifTypeLogProbabilityParameter);
		for(int k=0;k<motifModel.length;k++) {
			motifTypeLogProbabilityParameter[k] = Math.log(motifTypeLogProbabilityParameter[k]);
		}
	}

	public double getLogPriorTerm() throws Exception {
		return 0;
	}
	
	public double getLogProbFor(Sequence sequence, int startpos, int endpos) throws Exception {
		return Normalisation.logSumNormalisation(getUnnormalizedSequenceProfile(sequence.getSubSequence(startpos, endpos-startpos+1)));
	}
	
	public double getLogProbFor(DataSet data) throws Exception {
		double res = 0;
		for(int i=0;i<data.getNumberOfElements();i++) {
			res+=this.getLogProbFor(data.getElementAt(i),0,data.getElementAt(i).getLength()-1);
		}
		return res;
	}

	public String getInstanceName() {
		return "MultiOOPS";
	}

	public NumericalResultSet getNumericalCharacteristics() throws Exception {
		return null;
	}

	public boolean isInitialized() {
		return flankingModel.isInitialized()&&motifModel[0].isInitialized();
	}

	private double[] getUnnormalizedSequenceProfile(Sequence sequence) throws NotTrainedException, Exception {
		Sequence rc = null;
		if(strandOrientationLogProbabilityParameter.length!=1)
			rc = sequence.reverseComplement();
		double[] res = new double[motifModel.length*strandOrientationLogProbabilityParameter.length*(sequence.getLength()-motifModel[0].getLength()+1)];
			for(int i=0;i<sequence.getLength()-motifModel[0].getLength()+1;i++) {
				int p1 = Math.max(0,i-flankingModel.getMaximalMarkovOrder());
				int d1 = i-p1;
				int p2 = i+motifModel[0].getLength();
				int d2 = Math.min(sequence.getLength()-i-motifModel[0].getLength(),flankingModel.getMaximalMarkovOrder());
				int usedPositions = i*strandOrientationLogProbabilityParameter.length*motifModel.length;	
				double flank = - flankingModel.getLogProbFor(sequence,p1,p1+motifModel[0].getLength()+d1+d2-1) + flankingModel.getLogProbFor(sequence,p1,p1+d1-1) + flankingModel.getLogProbFor(sequence,p2,p2+d2-1);
				if(strandOrientationLogProbabilityParameter.length==1) {
					for(int k=0;k<motifModel.length;k++) {
						if(activeComponents[k])
							res[usedPositions+k] = flank+motifTypeLogProbabilityParameter[k]+motifModel[k].getLogProbFor(sequence,i,i+motifModel[k].getLength()-1);
						else
							res[usedPositions+k] = 0;
					}
				} else {
						for(int k=0;k<motifModel.length;k++) {
							if(activeComponents[k]) {
								res[usedPositions+2*k] = flank+ strandOrientationLogProbabilityParameter[0]+ motifModel[k].getLogProbFor(sequence,i,i+motifModel[k].getLength()-1);
								res[usedPositions+2*k+1] = flank+strandOrientationLogProbabilityParameter[1]+ motifTypeLogProbabilityParameter[k]+motifModel[k].getLogProbFor(rc,rc.getLength()-motifModel[k].getLength()-i,rc.getLength()-1-i);
							} else {
								res[usedPositions+2*k] = 0;
								res[usedPositions+2*k+1] = 0;
							}
						
					}
				}
			}
		return res;
	}
	
	/**
	 * Provides access to the {@link AbstractVariableStructureModel}s that are used as motif model.
	 * @return the motif models
	 */
	public AbstractVariableStructureModel[] getMotifModels() {
		return motifModel;
	}
	
	/**
	 * Provides access to the {@link AbstractTrainableStatisticalModel} that is used as flanking model.
	 * @return the flanking model
	 */
	public AbstractTrainableStatisticalModel getFlankingModel() {
		return flankingModel;
	}

	public String toString(NumberFormat nf) {
		return null;
	}
	
	private int getLV(int position, int type, int strand) {
		return position*motifModel.length*this.strandOrientationLogProbabilityParameter.length+type*this.strandOrientationLogProbabilityParameter.length+strand;
	}
	
	private int getPosition(int lv) {
		return (int)Math.floor(lv/(motifModel.length*this.strandOrientationLogProbabilityParameter.length));
	}
	
	private int getType(int lv) {
		return (int)Math.floor((lv%(this.strandOrientationLogProbabilityParameter.length*motifModel.length))/this.strandOrientationLogProbabilityParameter.length);
	}
	
	private int getStrand(int lv) {
		return (int)(lv%this.strandOrientationLogProbabilityParameter.length);
	}
	
	/**
	 * Allows to set a {@link Protocol} for writing output.
	 * @param protocol an instance of  {@link Protocol}
	 */
	public void setProtocol(Protocol protocol) {
		this.protocol=protocol;
		textOutput=true;
	}
	
	/**
	 * Allows to set a {@link ProgressUpdater}.
	 * @param progress an instance of  {@link  ProgressUpdater}
	 */
	public void setProgressUpdater(ProgressUpdater progress) {
		this.progress=progress;
	}
	
	/**
	 * Returns a log with information about all iteration steps during the optimization. 
	 * @return the log as {@link String}
	 */
	public String getLog() {
		return log.toString();
	}
	
	/**
	 * Checks whether a certain component is active or not. 
	 * @param k the component to be checked
	 * @return <code>true</code> if component <code>k</code> is active 
	 */
	public boolean checkComponentActivity(int k) {
		return activeComponents[k];
	}

	/**
	 * Enables or disables dynamic update of latent variables parameters during stochastic optimization.
	 * @param d true if dynamic update enabled 
	 */
	public void updateLVParam(boolean d) {
		updateLVParams = d;
	}
	
	/**
	 * Returns the optimal latent variable configuration, consisting of position, strand, and mixture component (type) of the motif  
	 * @return Optimal latent variable configuration as human-readable String
	 */
	public String getBestLVAssignment() {
		if(fullyObservable)
			return null;
		StringBuffer res = new StringBuffer();
		res.append("Seq.\t");
		if(!dataAligned)
			res.append("Pos.\t");
		if(strandOrientationLogProbabilityParameter.length>1)
			res.append("Strand\t");
		if(this.motifModel.length>1)
			res.append("Component");
		res.append("\n");
		for(int i=0;i<bestLV.length;i++) {
			res.append(i);
			res.append("\t");
			if(!dataAligned) {
				res.append(this.getPosition(bestLV[i]));
				res.append("\t");
			}
			if(strandOrientationLogProbabilityParameter.length>1) {
				if(this.getStrand(bestLV[i])==0) {
					res.append("+");
				} else {
					res.append("-");
				}
				res.append("\t");
			}			
			if(this.motifModel.length>1) {
				res.append(this.getType(bestLV[i]));
				res.append("\t");
			}
			res.append("\n");
		}
		return res.toString();
	}
	
	public String[] getBindingSites() {
		String[] s = new String[currentBindingSites.length]; 
		for(int c=0;c<currentBindingSites.length;c++) {
			StringBuffer out = new StringBuffer();
			for(int i=0;i<currentBindingSites[c].getNumberOfElements()-1;i++) {
				out.append(currentBindingSites[c].getElementAt(i)+"\n");
			}
			out.append(currentBindingSites[c].getElementAt(currentBindingSites[c].getNumberOfElements()-1));
			s[c] = out.toString();
		}
		return s;
	}

}
