/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.variableStructure.parsimonious;

import de.jstacs.data.AlphabetContainer;
import de.jstacs.io.NonParsableException;
import de.jstacs.projects.inmode.models.variableStructure.AbstractVariableStructureModel;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.dataStructures.ParsimoniousContextTree;


/**
 * Abstract class for all trainable statistical models that include {@link ParsimoniousContextTree}s. 
 * @author Ralf Eggeling
 *
 */
public abstract class AbstractParsimoniousModel extends AbstractVariableStructureModel {

	protected ParsimoniousContextTree[] tree;
		
	protected String[] parameterSamples;
	
	/** 
	 * This is the standard constructor.
	 * @param alphabets used alphabet container.
	 * @param length the length of the sequences to be handled (0 indicates arbitrary length).
	 */
	public AbstractParsimoniousModel(AlphabetContainer alphabets, int length) {
		super(alphabets, length);
	}
	/**
	 * The standard constructor for the interface {@link de.jstacs.Storable}. Creates a new {@link AbstractParsimoniousModel} out of a {@link StringBuffer}.
	 * @param sb the {@link StringBuffer} to be parsed
	 * @throws NonParsableException  if the {@link AbstractParsimoniousModel} could not be reconstructed out of
	 *             the {@link StringBuffer} <code>sb</code>      
	 */
	public AbstractParsimoniousModel(StringBuffer sb) throws NonParsableException {
		super(sb);
	}

	/**
	 * Returns a GraphViz representation of the PCTs.
	 * @return graphViz representation
	 */
	public String getGraphViz() {
		StringBuffer out = new StringBuffer();
		out.append("digraph{\n");
		out.append("edge [arrowsize=0.8 headport=n labelangle=0];\n");
		out.append("node  [label=\"\" width=1.2 height=0.55 color=black,fixedsize=true,fontsize=14];\n");
		for(int i=0;i<tree.length;i++) {
			out.append("/* PCT representation of position "+i+"*/\n");
			out.append(tree[i].getGraphviz());
		}
		out.append("}");
		return out.toString();
	}
	
	/**
	 * Returns a sparse parameter representation (one line) of the parsimonious model.
	 * @return the sparse parameter representation
	 */
	public String getSparseParameterRepresentation() {
		String res = "";
		for(int i=0;i<tree.length;i++) {
			res+=(tree[i].getSparseParameterRepresentation() + "&");
		}
		return res;
	}
	
	protected abstract void fromXML(StringBuffer xml) throws NonParsableException;
	
	/**
	 * Returns a structured parameter representation, where each line corresponds to exactly one leaf in a PCT. 
	 * @return the parameter representation
	 */
	public String getParameterRepresentation() {
		String res = "";
		for(int i=0;i<tree.length;i++) {
			res+="Position "+(i+1)+"\n";
			String[] str = tree[i].getSparseParameterRepresentation().split("%");
			for(int r=0;r<str.length;r++)
				res+=(str[r]+"\n");
		}
		return res;
	}
	
	
}