/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.variableStructure.parsimonious.dataStructures;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;

/**
 * The class implements a comparably array of binary values.
 * @author Ralf Eggeling
 *
 */
public class BinaryArray implements Comparable<BinaryArray> {

	private boolean[] arr;
	
	/**
	 * Standard constructor. Creates {@link BinaryArray} of given length with all fields initialized with <code>false</code>.
	 * @param length the length of the array
	 */
	public BinaryArray(int length) {
		arr = new boolean[length];
		Arrays.fill(arr, false);
	}
	
	/**
	 * Creates a {@link BinaryArray} object from a boolean array.
	 * @param arr the boolean array
	 */
	public BinaryArray(boolean[] arr) {
		this.arr = arr.clone();
	}
	
	/**
	 * Given a set of numbers, all fields indexed by the number are set to true.
	 * @param IDs a set of integers
	 */
	public void fillAccordingTo(HashSet<Integer> IDs) {
		Iterator<Integer> it = IDs.iterator();
		while(it.hasNext()) {
			arr[it.next()] = true;
		}
	}
	
	/**
	 * Returns the value at a given position.
	 * @param i position 
	 * @return value
	 */
	public boolean getVal(int i) {
		return arr[i];
	}
	
	/**
	 * Sets the value at a given position
	 * @param val value
	 * @param i position
	 */
	public void setVal(boolean val, int i) {
		arr[i]=val;
	}
	
	/**
	 * Returns the length of the array
	 * @return length
	 */
	public int getLength() {
		return arr.length;
	}
	
	/**
	 * Compares two {@link BinaryArray}s. Comparison is made according to the first non-equal field among both {@link BinaryArray}s, where a value of <code>true</code> preceeds <code>false</code>.
	 * @param o {@link BinaryArray} to compare to
	 * @return 1, if the instance is greater than <code>o</code>, -1 if smaller, 0 if equal   
	 */
	public int compareTo(BinaryArray o) {
		if(this.getLength()!=o.getLength())  {
			return this.getLength()-o.getLength();
		} else {
			for(int i=0;i<arr.length;i++) {
				if(arr[i]==true&&o.getVal(i)==false)
					return 1;
				if(arr[i]==false&&o.getVal(i)==true)
					return -1;
			}
			return 0;
		}
	}
	
	/**
	 * Checks whether two {@link BinaryArray}s are equal 
	 * @param x the {@link BinaryArray} to compare to 
	 * @return true, if equal
	 */
	public boolean equals(BinaryArray x) {
		if(this.compareTo(x)==0)
			return true;
		return false;
	} 
	
	public boolean equals(Object o) {
		return this.equals((BinaryArray)o);
	}
	
	public String toString() {
		StringBuffer s = new StringBuffer();
		for(int i=0;i<arr.length;i++)
			if(arr[i])
				s.append(1);
				else
					s.append(0);
		return s.toString();
	}
	
	public int hashCode() {
		return this.toString().hashCode();
	}
}
