/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.variableStructure.parsimonious.dataStructures;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;
import java.util.Vector;
import de.jstacs.NotTrainedException;
import de.jstacs.Storable;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.data.alphabets.DiscreteAlphabet;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.io.XMLParser;
import de.jstacs.projects.inmode.utils.NML;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.inhomogeneous.SequenceIterator;


/**
 * This class allows to learn a parsimonious context tree (PCT) given data and to use it for prediction. 
 * @author Ralf Eggeling
 *
 */
public class ParsimoniousContextTree implements Cloneable, Storable {
	
	boolean memoization = false;
	boolean pruning = false;
	

	private final String XML_TAG = "ParsimoniousContextTree";
	
	/**
	 * the root of the context tree
	 */
	public ContextTreeNode root;
	
	/**
	 * the maximal depth of the context tree
	 */
	private Byte maximalDepth;
	
	/**
	 * the data set used for training
	 */
	protected DataSet data;
	
	/**
	 * the weights of the data used for training
	 */
	protected double[] weights;
	
	/**
	 * the sum of the weights
	 */
	private double weightSum;
	
	/**
	 * the alphabet container of the full model
	 */
	private AlphabetContainer alphCon;

	/**
	 * structure of the current context tree as graphViz representation
	 */
	private String graphViz;
	
	/**
	 * probabilityParameters[a][b] is the probability of observing symbol b given context a. 
	 */
	private double[][] probabilityParameters;
	
	/**
	 * the regular expressions of all leaves of the context tree 
	 */
	private String[] regEx;
	
	/**
	 * the number of the ParsimoniousContextTree in a wider context, e.g. as part of an InhomogeneousParsimoniousMarkovModel.
	 * The PCTnumber indicates the random variable in the model that this ParsimoniousContextTree is responsible for.
	 */
	private int PCTnumber;
	
	/**
	 * next node to be visited during tree traversal
	 */
	private int nextNode;
	
	/**
	 * next leaf to be visited during tree traversal
	 */
	private int nextLeaf;
	
	/**
	 * structure of the current context tree as string
	 */
	private String treeStructure = "";
	
	
	/**
	 * Data structure for storing previously computer PCT subtrees or upper-bounded subtree scores
	 */
	public List<HashMap<BinaryArray,ContextTreeNode>> hml;
	
	/**
	 * All possible symbol sets (i.e. putative PCT nodes) at different levels of the PCT.
	 */
	public List<TreeSet<SymbolSet>> symbolsets;
	
	/**
	 * 
	 */
	int maxLookAhead = 1;
	
	/**
	 * true if the cond. prob. parameters are filled
	 */
	private boolean trained;
	
	/**
	 * Data structure for quick access to probability parameters
	 */
	HashMap<Sequence,Double> hmProbs;
	
	double[] condProbs;
	int[] condProbHelper;

	/**
	 * Standard constructor
	 * @param con The alphabet container
	 * @param maximalDepth  The maximal tree depth
	 * @throws Exception If something went wrong
	 */
	public ParsimoniousContextTree(AlphabetContainer con, Byte maximalDepth) throws Exception {
		this.maximalDepth=maximalDepth;
		this.alphCon=con;
		symbolsets = new ArrayList<TreeSet<SymbolSet>>();
		for(int i=0;i<maximalDepth;i++) {
			symbolsets.add(i, SymbolSet.createSymbolSets((DiscreteAlphabet) con.getAlphabetAt(i)));
		}
	}
	
	public ParsimoniousContextTree(StringBuffer xml) throws Exception {
		StringBuffer buff = XMLParser.extractForTag(xml, XML_TAG);
		this.alphCon = XMLParser.extractObjectForTags(buff, "AlphabetContainer",AlphabetContainer.class);	
		this.maximalDepth = XMLParser.extractObjectForTags(buff, "maximalDepth",Byte.class);
		this.trained = XMLParser.extractObjectForTags(buff,"trained",Boolean.class);
		if(trained) {
			this.treeStructure = XMLParser.extractObjectForTags(buff, "treeStructure", String.class);
			this.graphViz = XMLParser.extractObjectForTags(buff, "graphViz", String.class);
			this.regEx = XMLParser.extractObjectForTags(buff, "regularExpressions", String[].class);
			this.probabilityParameters = XMLParser.extractObjectForTags(buff, "probabilityParameters", double[][].class);
			this.condProbHelper = XMLParser.extractObjectForTags(buff, "condProbHelper", int[].class);
			this.condProbs = XMLParser.extractObjectForTags(buff, "condProbs", double[].class);
		}
	}
	
	
	public StringBuffer toXML() {
		StringBuffer xml = new StringBuffer();
		XMLParser.appendObjectWithTags(xml, alphCon, "AlphabetContainer");
		XMLParser.appendObjectWithTags(xml, maximalDepth, "maximalDepth");
		XMLParser.appendObjectWithTags(xml, trained, "trained");
		if(trained) {
			XMLParser.appendObjectWithTags(xml, treeStructure, "treeStructure");
			XMLParser.appendObjectWithTags(xml, graphViz, "graphViz");
			XMLParser.appendObjectWithTags(xml, regEx, "regularExpressions");
			XMLParser.appendObjectWithTags(xml, probabilityParameters, "probabilityParameters");
			XMLParser.appendObjectWithTags(xml, condProbs, "condProbs");
			XMLParser.appendObjectWithTags(xml, condProbHelper, "condProbHelper");
		}
		XMLParser.addTags(xml, XML_TAG);
		return xml;
	}
	
	public void maximize(DataSet data, double[] weights) {
		hml = new ArrayList<HashMap<BinaryArray,ContextTreeNode>>();
		for(int i=0;i<maximalDepth;i++) {
			hml.add(i,new HashMap<BinaryArray,ContextTreeNode>());
		}
		this.data=data;
		if(weights==null) {
			this.weights = new double[data.getNumberOfElements()];
			Arrays.fill(this.weights, 1);
		} else {
			this.weights=weights;
		}
		for(int i=0;i<this.weights.length;i++)  {
			weightSum+=this.weights[i];
		}
		root = new ContextTreeNode(null,null,maximalDepth);
		graphViz = "";
		probabilityParameters = new double[root.getNumberOfLeaves()][(int) alphCon.getAlphabetLengthAt(maximalDepth)];
		regEx = new String[root.getNumberOfLeaves()];
		nextNode=0;
		nextLeaf=0;
		root.traverseAfterConstruction();
		createCondProbTable();
		trained=true;
	}
	
	private void createCondProbTable() {
		hmProbs = new HashMap<Sequence,Double>();
		condProbs = new double[(int) Math.pow(DNAAlphabet.SINGLETON.length(),maximalDepth+1)];
		SequenceIterator it = new SequenceIterator(maximalDepth+1);
		if(maximalDepth>0) {
			condProbHelper = new int[maximalDepth];
			for(int s=0;s<maximalDepth;s++) {
				condProbHelper[s]= (int)Math.pow(DNAAlphabet.SINGLETON.length(),s+1);
			}
		}
		int[] bounds = new int[maximalDepth+1];
		Arrays.fill(bounds, (int) DNAAlphabet.SINGLETON.length());
		it.setBounds(bounds);
		int i=0;
		do {
			String str = "";
			for(int j=0;j<bounds.length;j++) {
				str+= alphCon.getSymbol(j, it.discreteValAt(j));
			} 
			Sequence seq = null;
			try {
				seq = Sequence.create(alphCon, str);
			} catch(Exception e) {
				e.printStackTrace();
			}
			condProbs[i]=Math.log(root.getProbFor(seq, maximalDepth));
			hmProbs.put(seq, condProbs[i]);
			i++;
		}
		while(it.next());
	}
	
	

	public int getNumberOfLeaves() {
		return probabilityParameters.length;
	}
	
	public double getScore() {
		double score = root.getLogScore();
/*		System.out.print(score+"\t");
		if(this.getActualDepth()==1) {
			score-=Math.log(15);
		}
		if(this.getActualDepth()==2) {
			score-=Math.log(72465);
		}
		if(this.getActualDepth()==3) {
			score-=Math.log(2.75E19);
		}
		if(this.getActualDepth()==4) {
			score-=Math.log(5.78E77);
		}
		if(this.getActualDepth()>=5) {
			score-=Math.log(1.79E308);
		} */
	//	System.out.println(score+"\t"+this.getActualDepth());
		return score;
	}
	
	public String getTreeStructure() {
		if(maximalDepth==0)
			return "()";
		else
			return "("+treeStructure.toString()+")";
	}
	

	/**
	 * 
	 * @param seq An arbitrary discrete sequence
	 * @param position the position of the symbol whose probability is to be computed
	 * @return returns the probability of the symbol at the given positions given the current context tree, given the precedent observations, given the probability parameters 
	 * @throws NotTrainedException If the model has not been trained before.
	 */
	public double getLogProbFor(Sequence seq, int position) throws NotTrainedException {
		if(!trained)
			throw new NotTrainedException();
		return condProbs[getIDofCondProb(seq,position-maximalDepth,position)];
	}
	
	public static double getOptimalPartitionScore(SymbolSet s, HashMap<String,SymbolSet> hm, double[] score) throws Exception {
		double res = Double.NEGATIVE_INFINITY;
		TreeSet<String> ts =  s.createSymbolMinimalSubSetsAsStrings();
		Iterator<String> it = ts.iterator();
		String best = "";
		while(it.hasNext()) {
			SymbolSet t = hm.get(it.next());
			String str = s.getSetDifferenceAsString(t);
			if (str.equals("") ) {
					if(s.getScore()>=res) {
						res = t.getScore();
						best = t.toSparseString();
					}
				} else {
					if(score[hm.get(str).ID]+t.getScore()>=res) {
						res = score[hm.get(str).ID]+t.getScore();
						best = t.toSparseString();
					}
				}
			}
		s.optimalSubSetIDs.clear();
		if(best.equals(s.toSparseString())) {
			s.optimalSubSetIDs.add(s.ID);
		} else {
			s.optimalSubSetIDs.add(hm.get(best).ID);
			s.optimalSubSetIDs.addAll(hm.get(s.getSetDifferenceAsString(hm.get(best))).optimalSubSetIDs);
		}
		return res;	
	}
	
	public void setRuntimeImprovements(boolean memoization, boolean pruning, int maxLookAhead) {
		this.memoization=memoization;
		this.pruning=pruning;
		this.maxLookAhead=maxLookAhead;
	}
	
	public void setRuntimeImprovements(boolean memoization, boolean pruning) {
		this.memoization=memoization;
		this.pruning=pruning;
		this.maxLookAhead=1;
	}

	public String getGraphviz() {
		return graphViz;
	}
	
	public void setTreeNumber(int i) {
		PCTnumber=i;
	}
	
	public ParsimoniousContextTree clone() throws CloneNotSupportedException {
		ParsimoniousContextTree cl = (ParsimoniousContextTree) super.clone();
		if(probabilityParameters!=null)
		cl.probabilityParameters = probabilityParameters.clone();
		if(regEx!=null)
		cl.regEx =regEx.clone();
		if(root!=null)
		cl.root = root.clone();
		return cl;
	}
	
	public String getSparseParameterRepresentation() {
		StringBuffer res = new StringBuffer();
		for(int i=0;i<probabilityParameters.length;i++) {
			res.append(regEx[i]);
			res.append("\t");
			for(int j=0;j<probabilityParameters[i].length;j++) {
				res.append(probabilityParameters[i][j]);
				res.append("\t");
			}
			res.append("%");
		}
		return res.toString();
	}

	
	public int getIDofCondProb(Sequence seq, int start, int end) {
		int res = 0;
		for(int d=end;d>start;d--) {
			res+= seq.discreteVal(d)*condProbHelper[d-start-1];
		}
		res+= seq.discreteVal(start);
		return res;
	}
	
	public int getActualDepth() {
		int usedLevels = 0;
		if(this.maximalDepth>0) {
			String[] struct = this.getSparseParameterRepresentation().split("%");
			for(int i=0;i<struct.length;i++) {
				struct[i] = struct[i].split("\t")[0];
	//			System.out.println(i+"\t"+struct[i]);
				int pru = 0;
				boolean nosplit = true;
				while((struct[i].length()>5)&&nosplit) {
					if(struct[i].substring(0, 6).equals("[ACGT]")) {
						pru++;
						struct[i] = struct[i].substring(6);
					} else {
						nosplit=false;
					}
				}
				if(maximalDepth-pru>usedLevels) 
					usedLevels=maximalDepth-pru;
			}
		}
		return usedLevels;
	}

	private class ContextTreeNode implements Comparable<ContextTreeNode>, Cloneable {

		/**
		 * the distance of this node to the root
		 */
		private int distanceToRoot;
		
		/**
		 * the label of the node, which is a subset of all possible symbols at this layer of the context tree
		 */
		private final SymbolSet label;
		
		/**
		 * the regular expression encoded by this node
		 */
		private String regExStr;
		
		/**
		 * the number of leaves of the subtree under this node
		 */
		private int numberOfLeaves = 0;
		
		/**
		 * the log score of the context tree under this leaf
		 */
		private double logScore;
		
		/**
		 * the set of children nodes, null if node is leaf
		 */
		private TreeSet<ContextTreeNode> children;
		
		/**
		 * The number of the node in the final context tree
		 */
		protected int nodeNumber = 0;
		
		/**
		 * the index of the leaf corresponding to the order of being visited during tree traversal
		 */
		private int leafIndex;
		
		/**
		 * the counts of the response variable for all contexts matching this this node
		 */
		private double[] counts;
		
		/**
		 *	Indicates the IDs of the sequences matching this node.
		 */
		protected HashSet<Integer> representedSequenceIDs;
		
		/**
		 *	Indicates the IDs of the sequences matching this node.
		 */
		protected Vector<HashSet<Integer>> representedSequenceIDsNextVariable;
		
		/**
		 * Represents labels and scores of putative children
		 */
		TreeSet<SymbolSet> childNodeLabels;
		
		
		protected int lookAheadStatus;
		
		protected ContextTreeNode(ContextTreeNode parent) {
			distanceToRoot = parent.getDistanceToRoot() + 1;
			label = symbolsets.get(distanceToRoot-1).last();
			logScore=parent.logScore;
			String str = "[";
			for(int i=0;i<label.getNumberOfElements();i++) {
				str+=label.getElements()[i];
			}
			str+="]";
			regExStr = str+parent.getRegExStr();
			numberOfLeaves=1;
			counts = parent.counts;
			children = new TreeSet<ContextTreeNode>();
			if(distanceToRoot<maximalDepth) {
				children.add(new ContextTreeNode(this));
			}
			
		}
		
		protected ContextTreeNode(ContextTreeNode parent, SymbolSet label, int lookAheadLevel) {
			// data management
			representedSequenceIDsNextVariable = new Vector<HashSet<Integer>>();
			for(int c=0;c<data.getAlphabetContainer().getMaximalAlphabetLength();c++) {
				representedSequenceIDsNextVariable.add(new HashSet<Integer>());
			}
					
			if(parent==null &&label==null) {
				this.label=null;
				distanceToRoot=0;
				regExStr = "";
				representedSequenceIDs = new HashSet<Integer>();
				for(Integer i=0;i<data.getNumberOfElements();i++) {
					representedSequenceIDs.add((Integer)i);
				}
				if(maximalDepth>0)
				for(Integer i=0;i<data.getNumberOfElements();i++) {
					int s = data.getElementAt(i).discreteVal(data.getElementAt(i).getLength()-2);
					representedSequenceIDsNextVariable.get(s).add(i);					
				}
			} else {
				this.label = label;
				distanceToRoot = parent.getDistanceToRoot() + 1;
				String str = "[";
				for(int i=0;i<label.getNumberOfElements();i++) {
					str+=label.getElements()[i];
				}
				str+="]";
				regExStr = str+parent.getRegExStr();
				if(label.getNumberOfElements()==alphCon.getAlphabetLengthAt(0)) {
					representedSequenceIDs = parent.representedSequenceIDs;
					if(this.distanceToRoot<maximalDepth) {
						Iterator<Integer> it = representedSequenceIDs.iterator();
						while(it.hasNext()) {
							Integer id = it.next();
							int s = data.getElementAt(id).discreteVal(data.getElementAt(id).getLength()-2-this.distanceToRoot);
							representedSequenceIDsNextVariable.get(s).add(id);		
						}
					}
				} else {
					representedSequenceIDs = new HashSet<Integer>();
					int[] symbolsCode = label.getCode();	
					for(int s=0;s<symbolsCode.length;s++) {
						representedSequenceIDs.addAll(parent.representedSequenceIDsNextVariable.elementAt(symbolsCode[s]));
					}
					if(this.distanceToRoot<maximalDepth) {
						Iterator<Integer> it = representedSequenceIDs.iterator();
						while(it.hasNext()) {
							Integer id = it.next();
							int s = data.getElementAt(id).discreteVal(data.getElementAt(id).getLength()-2-this.distanceToRoot);
							representedSequenceIDsNextVariable.get(s).add(id);		
						}
					}
				}
			}
			counts = computeCounts();
			
			
			// actual work
			if(pruning) {
				if(distanceToRoot==lookAheadLevel) {
					//score computation
					if(distanceToRoot==maximalDepth) {
						logScore = this.computeLogBICScoreOfLeaf();
						this.numberOfLeaves = 1;
					} else {
						logScore = Math.max(this.getLogLikelihoodTermOfDistinctSequenceSplit(representedSequenceIDs, this.distanceToRoot)-2*(alphCon.getAlphabetLengthAt(0)-1)*Math.log(data.getNumberOfElements())/2,this.computeLogBICScoreOfLeaf());
					}
				} else {
					if(this.getLogLikelihoodTermOfDistinctSequenceSplit(representedSequenceIDs, this.distanceToRoot)-2*(alphCon.getAlphabetLengthAt(0)-1)*Math.log(data.getNumberOfElements())/2<=this.computeLogBICScoreOfLeaf()) {
						this.logScore=this.computeLogBICScoreOfLeaf();
						this.numberOfLeaves=1;
						this.children = new TreeSet<ContextTreeNode>();
						this.children.add(new ContextTreeNode(this));
					} else  {
						int numberOfLookAheads = Math.min(maxLookAhead+1,lookAheadLevel-this.distanceToRoot); 
						for(int p=0;p<numberOfLookAheads;p++) {
							int j=this.distanceToRoot+1+p;
						if((p==numberOfLookAheads-1))
							j=lookAheadLevel;
						// memoization
						BinaryArray bn = null;
						boolean memoActive = false;
						if(memoization) {
							bn = new BinaryArray(data.getNumberOfElements());
							bn.fillAccordingTo(representedSequenceIDs);
							if(hml.get(distanceToRoot).containsKey(bn)) {
								ContextTreeNode cn = hml.get(distanceToRoot).get(bn);
								if(cn.lookAheadStatus>=j) {
									this.lookAheadStatus=cn.lookAheadStatus;
									this.logScore=cn.logScore;
										if(cn.lookAheadStatus==maximalDepth) {
											this.numberOfLeaves = cn.numberOfLeaves;
											this.children = (TreeSet<ContextTreeNode>) cn.children.clone(); 
										}
										memoActive=true;
										p=numberOfLookAheads;
									}
								}
							}
						if(memoActive==false)  {
							if(childNodeLabels==null) {
								try{			
								childNodeLabels = SymbolSet.createSymbolSets((DiscreteAlphabet) alphCon.getAlphabetAt(maximalDepth-1-distanceToRoot));
								} catch (Exception e) {
									e.printStackTrace();
								}
								children = new TreeSet<ContextTreeNode>();
								ContextTreeNode child = new ContextTreeNode(this,childNodeLabels.last(),lookAheadLevel);
								childNodeLabels.last().setScore(child.logScore);
								children.add(child);
							}
							// compute (scores of) child nodes (recursion)
							for(SymbolSet set: childNodeLabels) {
								if(set.isActiveChild && set.size()<alphCon.getMaximalAlphabetLength()) {
									ContextTreeNode child = new ContextTreeNode(this,set,j);
									if(children.contains(child)) {
										children.remove(child);
									}
									children.add(child);
									set.setScore(child.logScore);
								}
							}
							
							
							// compute partition scores
							double[] partitionScore = new double[ childNodeLabels.size()];
							Arrays.fill(partitionScore, 0);
							SymbolSet[] ssarr = new SymbolSet[  childNodeLabels.size()];
							HashMap<String,SymbolSet> hm = new HashMap<String,SymbolSet>();
							int k=0;
							Iterator<SymbolSet> it =  childNodeLabels.iterator();
							while(it.hasNext()) {
								SymbolSet s = it.next();
								hm.put(s.toSparseString(), s);
								ssarr[k] = s;
								s.ID=k;
								k++;
							}
							for(int i=0;i<ssarr.length;i++) {
								try {
									partitionScore[i] = getOptimalPartitionScore(ssarr[i],hm,partitionScore);
								}
								catch(Exception e) {
									partitionScore[i] = Double.NEGATIVE_INFINITY;
									System.out.println("Exception at score computation.");
								}
							}
							this.logScore=partitionScore[ssarr.length-1];
							if(this.distanceToRoot>0)
							this.label.setScore(this.logScore);
							if(j<maximalDepth) {
								// Stopping rule
								if(partitionScore[partitionScore.length-1]<=this.computeLogBICScoreOfLeaf()) {
									this.logScore=this.computeLogBICScoreOfLeaf();
									this.numberOfLeaves=1;
									this.children.clear();
									this.children.add(new ContextTreeNode(this));
									p=numberOfLookAheads+1;
								}  else {
								// Deletion rule
									for(int a=0;a<partitionScore.length-1;a++) {
										for(int b=0;b<partitionScore.length-1;b++) {
											if(!ssarr[a].containsAtLeastOneElementOf(ssarr[b])&&ssarr[a].isActiveChild()&&ssarr[b].isActiveChild()) {
												TreeSet<String> tsloc= new TreeSet<String>();
												tsloc.addAll(ssarr[a].getSymbolsAsSet());
												tsloc.addAll(ssarr[b].getSymbolsAsSet());
												if(tsloc.equals(ssarr[ssarr.length-1].getSymbolsAsSet())) {
													if(ssarr[ssarr.length-1].getScore()>(ssarr[a].getScore() + partitionScore[b])) {
														ssarr[a].disable();
													}
												}
											}
										}
									} 
								}
							} else {
								TreeSet<Integer> tsi = ssarr[ssarr.length-1].optimalSubSetIDs;
								TreeSet<SymbolSet> tobeRetained = new TreeSet<SymbolSet>();
								while(!tsi.isEmpty()) {
									SymbolSet s = ssarr[tsi.pollFirst()];
									tobeRetained.add(s);
								}
								TreeSet<ContextTreeNode> toBeRemoved = new TreeSet<ContextTreeNode>();
								for(ContextTreeNode tau : this.children) {
									if(!tobeRetained.contains(tau.getLabel())) {
										toBeRemoved.add(tau);
									}
								}
								for(ContextTreeNode tau : toBeRemoved) {
									children.remove(tau);
								}
								toBeRemoved.clear();
								
													
								// update score and leaf number of this node 
								for(ContextTreeNode tau: this.children) {
									this.numberOfLeaves += tau.getNumberOfLeaves();
								}
								this.lookAheadStatus=j;
								if(memoization) {
									hml.get(distanceToRoot).remove(bn);
									hml.get(distanceToRoot).put(bn, this);
								}
							}
						}
						
					}
				}
				}
			} else {
				if(distanceToRoot<maximalDepth) {
					boolean memoActive = false;
					BinaryArray bn = null;
					if(memoization) {
						bn = new BinaryArray(data.getNumberOfElements());
						bn.fillAccordingTo(representedSequenceIDs);
						if(hml.get(distanceToRoot).containsKey(bn)) {
							ContextTreeNode cn = hml.get(distanceToRoot).get(bn);
							this.children = (TreeSet<ContextTreeNode>) cn.children.clone(); 
							this.logScore = cn.logScore;
							this.numberOfLeaves = cn.numberOfLeaves;
							memoActive=true;
						}
					}
					if(memoActive==false)   {
						try{
							childNodeLabels = SymbolSet.createSymbolSets((DiscreteAlphabet) alphCon.getAlphabetAt(maximalDepth-1-distanceToRoot));
						} catch (Exception e) {
							e.printStackTrace();
							}
						children = new TreeSet<ContextTreeNode>();
						for(SymbolSet set: childNodeLabels) {
								ContextTreeNode child = new ContextTreeNode(this,set,0);
								children.add(child);
								set.setScore(child.logScore);
							}
						
						double[] gScore = new double[ childNodeLabels.size()];
						Arrays.fill(gScore, 0);
						SymbolSet[] ssarr = new SymbolSet[  childNodeLabels.size()];
						HashMap<String,SymbolSet> hm = new HashMap<String,SymbolSet>();
						int k=0;
						Iterator<SymbolSet> it =  childNodeLabels.iterator();
						while(it.hasNext()) {
							SymbolSet s = it.next();
							hm.put(s.toSparseString(), s);
							ssarr[k] = s;
							s.ID=k;
							k++;
						}
						for(int i=0;i<ssarr.length;i++) {
							try {
								gScore[i] = getOptimalPartitionScore(ssarr[i],hm,gScore);
							}
							catch(Exception e) {
								gScore[i] = Double.NEGATIVE_INFINITY;
								System.out.println("Exception at score computation.");
							}
						}
						TreeSet<Integer> tsi = ssarr[ssarr.length-1].optimalSubSetIDs;
						TreeSet<SymbolSet> tobeRetained = new TreeSet<SymbolSet>();
						while(!tsi.isEmpty()) {
							SymbolSet s = ssarr[tsi.pollFirst()];
							tobeRetained.add(s);
						}
				
						// remove all children not belonging to the maximal
						TreeSet<ContextTreeNode> toBeRemoved = new TreeSet<ContextTreeNode>();
						for(ContextTreeNode tau : this.children) {
							if(!tobeRetained.contains(tau.getLabel())) {
								toBeRemoved.add(tau);
							}
						}
						for(ContextTreeNode tau : toBeRemoved) {
							children.remove(tau);
						}
						toBeRemoved.clear();
											
						// update score and leaf number of this node 
						for(ContextTreeNode tau: this.children) {
							this.logScore += tau.getLogScore();
							this.numberOfLeaves += tau.getNumberOfLeaves();
						}		
						if(memoization) {
							hml.get(distanceToRoot).put(bn, this);
						}
					}
					
				} else {
					numberOfLeaves = 1;
					logScore = this.computeLogBICScoreOfLeaf();
				}

			}

		}
		
		
		/**
		 * A node A is smaller than a node B if the label/SymbolSet of A is smaller than the label/SymbolSet of B 
		 */
		public int compareTo(ContextTreeNode o) {
			return this.label.compareTo(o.label);
		}
		
		/**
		 * 
		 * @return score of the tree
		 */
		protected double getLogScore() {
			return logScore;
		}
		
		
		/**
		 * if return value 0 -> node is root
		 * if return value maximalDepth -> node is leaf
		 * @return distance to root
		 */
		protected int getDistanceToRoot() {
			return distanceToRoot;
		}
		
		/**
		 * 
		 * @return the label/SymbolSet of a node
		 */
		protected SymbolSet getLabel() {
			return label;
		}
		
		/**
		 * 
		 * @return the number of leaves of the subtree under this node
		 */
		protected int getNumberOfLeaves() {
			return numberOfLeaves;
		}
		
		/**
		 * 
		 * @return the regular expression encoded by this node
		 */
		protected String getRegExStr() {
			return regExStr;
		}
		
		/**
		 * traverses the tree after construction. computes compact tree representation and computes conditional probability parameters. 
		 */
		protected void traverseAfterConstruction() {
			String l = "";
			if(label!=null) {
				l=this.label.toString().substring(1,this.label.toString().length()-1);
			}				
			graphViz+=("t"+PCTnumber+"n"+this.nodeNumber+"[label=\""+l+"\"];\n");
			if(distanceToRoot<maximalDepth) {
				Iterator<ContextTreeNode> iter = children.iterator();
				while(iter.hasNext()) {
					nextNode++;
					ContextTreeNode n = iter.next();
					n.nodeNumber=nextNode;
					graphViz+=("t"+PCTnumber+"n"+this.nodeNumber+"-> t"+PCTnumber+"n"+n.nodeNumber+";\n");
				}
				if(label!=null) {
					treeStructure+=(label+"(");
				} 
				Iterator<ContextTreeNode> it = children.iterator();
				while(it.hasNext()) {
					it.next().traverseAfterConstruction();
				}
				if(label!=null)
				treeStructure+=")";
			} else {
				treeStructure+=(label);
				{
					leafIndex = nextLeaf;
						double sum = 0;
						for(int i=0;i<counts.length;i++) {
							sum += NML.fsNMLeTerm((int)counts[i])*(counts[i]+1);
						}
						for(int i=0;i<counts.length;i++) {
							probabilityParameters[leafIndex][i] = NML.fsNMLeTerm((int)counts[i])*(counts[i]+1)/sum;
						}
					regEx[leafIndex] = regExStr;
					nextLeaf++;
				}
				
			}
		}
		
		protected double[] computeCounts() {
		double[] lcounts = new double[(int)alphCon.getAlphabetAt(alphCon.getNumberOfAlphabets()-1).length()];
		Iterator<Integer> it = representedSequenceIDs.iterator();
		while(it.hasNext()) {
				int next = it.next();
				lcounts[data.getElementAt(next).discreteVal(data.getElementAt(next).getLength()-1)]+=weights[next];
			}
		return lcounts;
		}
		
		public double getLogLikelihoodTermOfDistinctSequenceSplit(HashSet<Integer> representedSequenceIDs, int distanceToRoot) {
			double res = 0;
			Iterator<Integer> its = representedSequenceIDs.iterator();
			TreeSet<String> tss = new TreeSet<String>();
			HashMap<String,double[]> hm = new HashMap<String,double[]>();
			while(its.hasNext()) {
				int locID = its.next();
				String context = data.getElementAt(locID).getSubSequence(0, data.getElementAt(locID).getLength()-1-distanceToRoot).toString();
				double[] localCounts;
				if(hm.containsKey(context)) {
					localCounts = hm.get(context);
				} else {
					tss.add(context);			
					localCounts = new double[(int)alphCon.getAlphabetAt(alphCon.getNumberOfAlphabets()-1).length()];
				}
				localCounts[data.getElementAt(locID).discreteVal(data.getElementAt(locID).getLength()-1)]+=weights[locID];
				hm.put(context, localCounts);
			}
			Iterator<String> ittss = tss.iterator();
			while(ittss.hasNext()) {
				double[] localCounts = hm.get(ittss.next());
				double localN = 0;
				for(int i=0;i<localCounts.length;i++) {
					localN += localCounts[i];
				}
				double logLikelihoodTerm = 0;
				for(int i=0;i<localCounts.length;i++) {
					if(localCounts[i]>0)
						logLikelihoodTerm += localCounts[i]*(Math.log(localCounts[i])-Math.log(localN));
				}
				res+=logLikelihoodTerm;
			}
			return res;
		}
		
		protected double computeLogBICScoreOfLeaf() {
			double localN = 0;
			for(int i=0;i<counts.length;i++) {
				localN += counts[i];
			}
			double logLikelihoodTerm = 0;
			for(int i=0;i<counts.length;i++) {
				if(counts[i]>0)
				logLikelihoodTerm += counts[i]*(Math.log(counts[i])-Math.log(localN));
			}
			return logLikelihoodTerm - (counts.length-1)*Math.log(weightSum)/2;
						
		}
		
		protected double getProbFor(Sequence seq, int position) {
			double res = 0;
			if(distanceToRoot<maximalDepth) {
				Iterator<ContextTreeNode> it = children.iterator();
				while(it.hasNext()) {
					ContextTreeNode child = it.next();
					if(child.label.contains(
					((DiscreteAlphabet)alphCon.getAlphabetAt(maximalDepth-1-distanceToRoot)).getSymbolAt( seq.discreteVal(position-1-distanceToRoot )))) {
						res =  child.getProbFor(seq, position);
						break;
					}
				}
			} else {
				res = probabilityParameters[leafIndex][seq.discreteVal(position)];
			} 
			return res;
		}
		
		public ContextTreeNode clone() throws CloneNotSupportedException {
			ContextTreeNode cl = (ContextTreeNode)super.clone();
			return cl;
		}
		
	}
}
