/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.variableStructure.parsimonious.dataStructures;
import java.util.Iterator;
import java.util.TreeSet;

import de.jstacs.data.alphabets.DiscreteAlphabet;


/**
 * Represents a partition of a given {@link DiscreteAlphabet}.
 * @author Ralf Eggeling
 *
 */
public class Partition implements Comparable<Partition> {
	
	private double score;
	
	private DiscreteAlphabet alph;
	
	private TreeSet<SymbolSet> set;
	
	private int numberOfSymbols = 0;
	
	private int numberOfMergedNodes = 0;
	
	protected Partition(DiscreteAlphabet alph, SymbolSet...sets ) {
		set = new TreeSet<SymbolSet>();
		this.alph = alph;
		for(int i=0;i<sets.length;i++) {
			set.add(sets[i]);
			numberOfSymbols+=sets[i].getNumberOfElements();
			if(sets[i].getNumberOfElements()>1) {
				numberOfMergedNodes++;
			}
		}
	}
	
	protected void add(SymbolSet s) {
		set.add(s);
		numberOfSymbols+=s.getNumberOfElements();
	}
	
	public String toString() {
		return set.toString();
	}
	
	protected Boolean equals(Partition p) {
		if(alph.equals(p.getAlphabet())) {
//			System.out.println("jooo");
			return set.equals(p.set);
		} else {
			return false;
		}
	}
	
	protected DiscreteAlphabet getAlphabet() {
		return alph;
	}
	
	/**
	 * 
	 * @return true if each symbol of the alphabet occurs in exactly one of the SymbolSets of the Partition
	 */
	protected Boolean isValidPartition() {
		try {
			Object[] arr = set.toArray();
			SymbolSet[] sysearr = new SymbolSet[arr.length];
			int numberOfElements = 0;
			for(int i=0;i<arr.length;i++) {
				sysearr[i]=(SymbolSet)arr[i];
				numberOfElements+=sysearr[i].getNumberOfElements();
			}
			if(numberOfElements!=alph.length()) {
				return false;
			} else {
				return new DiscreteAlphabet(true,(new SymbolSet(alph,false)).union(sysearr).getElements()).checkConsistency(alph);
			}
		}catch(Exception e) {
			return false;
		}
	}
	
	protected Boolean isValidPartitionUnderSizeConstraint(int allowedSize, int allowedNumberOfExceptions) {
		if(!this.isValidPartition())
			return false;
		else {
			int exceptions=0;
			Iterator<SymbolSet> it = set.iterator();
			while(it.hasNext()) {
				if(it.next().size()>allowedSize) {
					exceptions++;
				}
				if(exceptions>allowedNumberOfExceptions) {
		//			System.out.println("Partition "+this.toString()+"\trejected");
					return false;
					
				}
			}
		}
	//	System.out.println("Partition "+this.toString()+"\taccepted");
		return true;
	}
	
	protected void setScore(double score) {
		this.score=score;
	}
	
	protected double getScore() {
		return this.score;
	}
	
	protected int getNumberOfSymbols() {
		return numberOfSymbols;
	}
	
	/**
	 * Checks if the Partition could be extended by the given SymbolSet.
	 * @param s an arbitrary SymbolSet
	 * @return true, if the number of symbols in the partition plus the number of symbols in s 
	 * is less or equal to the alphabet size and if none of the symbols in s is already contained 
	 * in of the SymbolSets of the partition
	 */
	protected boolean isExtendableBy(SymbolSet s) {
		if(this.getNumberOfSymbols()+s.getNumberOfElements()>alph.length())
			return false;
		Object[] arr = set.toArray();
		SymbolSet[] sysearr = new SymbolSet[arr.length];
		for(int i=0;i<arr.length;i++) {
			sysearr[i]=(SymbolSet)arr[i];
			if(sysearr[i].containsAtLeastOneElementOf(s))
				return false;
		}
		return true;
	}
	/**
	 * An order among Partitions is defined according to following rules:
	 * 1. number of elements (a Partition p is smaller than a Partition p' if p contains more elements than p')
	 * 2. comparing the elements is ascending order
	 * Two Partitions are equal if and only if they contain pairwise equal SymbolSets (not necessarily the same objects)
	 */
	
	public int compareTo(Partition o) {
		if(o.set.size()!=this.set.size()) {
	//		System.out.println("Partition number of elements do not match");
			return (int)Math.signum(o.set.size()-this.set.size());
		}
	//	System.out.println("Partition number of elements match");
		Iterator<SymbolSet> ownIt = this.set.iterator();
		Iterator<SymbolSet> foreignIt = o.set.iterator();
		for(int i=0;i<Math.min(o.set.size() ,this.set.size() );i++) {
			SymbolSet ownSymbolSet = ownIt.next();
			SymbolSet foreignSymbolSet = foreignIt.next();
			boolean equal = foreignSymbolSet.equals(ownSymbolSet);
	//		System.out.println(i+"\t"+equal);
			if(!equal)
				return ownSymbolSet.compareTo(foreignSymbolSet);
		}
	//	System.out.println("Return zero as default");
		return 0;
	}
	
	/**
	 * 
	 * @param s an arbitrary SymbolSet
	 * @return true if the partition contains s
	 */
	protected boolean contains(SymbolSet s) {
		for(SymbolSet t:set) {
			if(s.compareTo(t)==0)
				return true;
		}
		return false;
	}
	
	/**
	 * Returns the number of merged nodes (i.e. subsets of size>1) within this partition
	 * @return The number of subsets of size>1 in this partition.
	 */
	protected int getNumberOfMergedNodes() {
		return numberOfMergedNodes;
	}

}