/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous;

import de.jstacs.DataType;
import de.jstacs.io.NonParsableException;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.io.XMLParser;
import de.jstacs.parameters.SequenceScoringParameterSet;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.validation.NumberValidator;

/**
 * ParameterSet for {@link InhomogeneousPMM}. 
 * @author Ralf Eggeling
 *
 */
public class InhPMMParameterSet extends SequenceScoringParameterSet<InhomogeneousPMM> {

	
	
	protected static final String XMLTAG="InhPMMParameterSet";
	
	
	/**
	 * This is the standard Constructor
	 * @param con the alphabet container
	 * @param length the length of the sequences to be scored
	 * @param order the maximal/initial model order
	 * @throws WrongAlphabetException if the alphabet is continuous 
	 */
	public InhPMMParameterSet(AlphabetContainer con, int length, byte order)
			throws Exception {
		super(InhomogeneousPMM.class, con, length, false);
		if(!con.isDiscrete())
			throw new WrongAlphabetException("The model only supports discrete alphabets.");
		parameters.add(new SimpleParameter(DataType.BYTE,"Order","The maximal/initial order of the iPMM",true,new NumberValidator<Byte>((byte)0,Byte.MAX_VALUE),(byte)2));
		parameters.get(0).setValue(order);
	}
	
	/**
	 * The standard constructor for the interface {@link de.jstacs.Storable}.
	 * Creates a new {@link InhPMMParameterSet} out of a {@link StringBuffer}.
	 * 
	 * @param sb
	 *            the {@link StringBuffer} to be parsed
	 * 
	 * @throws NonParsableException
	 *             is thrown if the {@link StringBuffer} could not be parsed
	 */	
	public InhPMMParameterSet(StringBuffer sb) throws NonParsableException {
		super(sb);
	}

	public String getInstanceComment() {
		return "The ParameterSet <code>InhPMMPParameterSet</code> is used to initialize <code>InhomogeneousPMM</code>.";
	}

	public String getInstanceName() {
		return "InhomogeneousParsimoniousMarkovModel";
	}
	
	public String getXMLTAG	(){
		return XMLTAG;
	}
	
	public StringBuffer toXML() {
		StringBuffer sb = super.toXML();
		XMLParser.addTags(sb, XMLTAG);
		return sb;
	}
		
	public void fromXML ( StringBuffer sb ) throws NonParsableException {
		super.fromXML(XMLParser.extractForTag(sb,getXMLTAG()));
	}

	
	public InhPMMParameterSet clone() throws CloneNotSupportedException  {
		InhPMMParameterSet cl = (InhPMMParameterSet)super.clone();
		return cl;
	}

}