/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous;

import java.text.NumberFormat;
import java.util.Arrays;

import de.jstacs.InstantiableFromParameterSet;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongLengthException;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.io.NonParsableException;
import de.jstacs.io.XMLParser;
import de.jstacs.parameters.InstanceParameterSet;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.AbstractParsimoniousModel;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.dataStructures.ParsimoniousContextTree;
import de.jstacs.results.NumericalResult;
import de.jstacs.results.NumericalResultSet;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.inhomogeneous.SequenceIterator;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.utils.Normalisation;


/**
 * This class represents an inhomogeneous parsimonious Markov model (iPMM).
 * @author Ralf Eggeling
 *
 */
public class InhomogeneousPMM extends AbstractParsimoniousModel
		implements InstantiableFromParameterSet {
	
	protected static final String XML_TAG = "InhomogeneousPMM";

	/**
	 * the parameter set of the full model
	 */
	protected InhPMMParameterSet params;
	
	private byte order;
	
	private boolean memoization = false;
	private boolean pruning = false;
	
	private ProgressUpdater progress;
		
	/**
	 * Constructs an inhomogeneous parsimonious Markov model (PMM) from a 
	 * @param params given parameter set.
	 * @throws Exception if something went wrong
	 */
	public InhomogeneousPMM(InhPMMParameterSet params) throws Exception {
		super(params.getAlphabetContainer(),params.getLength());
		try{
			this.params = params.clone();
		} catch(CloneNotSupportedException cnse) {
			System.out.println("ParameterSet could not be cloned. Use reference instead");
			this.params=params;
		}
		tree = new ParsimoniousContextTree[params.getLength()];
		order = (Byte)params.getParameterAt(0).getValue();
	}
	
	
	/**
	 * The standard constructor for the interface {@link de.jstacs.Storable}.
	 * Creates a new {@link InhomogeneousPMM} out of a {@link StringBuffer}.
	 * 
	 * @param sb
	 *            the {@link StringBuffer} to be parsed
	 * 
	 * @throws NonParsableException
	 *             is thrown if the {@link StringBuffer} could not be parsed
	 */	
	public InhomogeneousPMM(StringBuffer sb) throws NonParsableException {
		super(XMLParser.extractForTag(sb,XML_TAG));
	}

	protected void fromXML(StringBuffer xml) throws NonParsableException {
		this.tree = XMLParser.extractObjectForTags(xml,"ParsimoniousContextTrees",ParsimoniousContextTree[].class);
		this.params = XMLParser.extractObjectForTags(xml, "params", InhPMMParameterSet.class);
		this.trained = XMLParser.extractObjectForTags(xml, "trained", boolean.class);
		alphabets = params.getAlphabetContainer();
		length = params.getLength();
		order=(Byte)params.getParameterAt(0).getValue();
		}

	
	public StringBuffer toXML() {
		StringBuffer xml = new StringBuffer();
		XMLParser.appendObjectWithTags(xml, params, "params");
		XMLParser.appendObjectWithTags(xml, tree, "ParsimoniousContextTrees");
		XMLParser.appendObjectWithTags(xml, trained, "trained");
		XMLParser.addTags(xml, XML_TAG);
		return xml;
	}

	public String getInstanceName() {
		return "iPMM("+params.getParameterAt(0).getValue()+")";
	}

	public double getLogPriorTerm() throws Exception {
		return 0;
	}

	public NumericalResultSet getNumericalCharacteristics() throws Exception {
		return new NumericalResultSet(new NumericalResult("numberOfLeaves","The total number of leaves of the inh. PMM",this.getNumberOfLeaves()));
	}

	public double getLogProbFor(Sequence seq, int start, int end) throws Exception {
		if(end-start+1!=length)
			throw new WrongLengthException("Sequence and model length do not match: "+(end-start+1)+"!="+length);
		double res = 0;
		for(int i=0;i<=end-start;i++) {
			double p = tree[i].getLogProbFor(seq, i+start);
	//		System.out.println(p);
			res += p;
		}
		return res;
	}
	
	public String getXMLTag() {
		return XML_TAG;
	}
	
	public InhomogeneousPMM clone() throws CloneNotSupportedException {
		InhomogeneousPMM cl = (InhomogeneousPMM)super.clone();
		cl.params = params.clone();
		cl.tree = new ParsimoniousContextTree[tree.length];
		for(int i=0;i<tree.length;i++) {
			if(tree[i]!=null)
				cl.tree[i] = tree[i].clone();
		}
		cl.trained=trained;
		return cl; 
	}
	
	/**
	 * Returns the model complexity in terms of the number of all PCT leaves at all sequence positions.
	 * @return the number of leaves
	 */
	public int getNumberOfLeaves() {
		int res = 0;
		for(int i=0;i<tree.length;i++) {
			res+= tree[i].getNumberOfLeaves();
		}
		return res;
	}

	@Override
	public void train(DataSet data, double[] weights) throws Exception {
		if(weights == null) {
			weights = new double[data.getNumberOfElements()];
			Arrays.fill(weights,1);
		}
		for(int i=0;i<tree.length;i++) {
				DataSet dataForThisTree = data.getInfixDataSet(i-Math.min(i,order),1+Math.min(i,order));
				tree[i]=new ParsimoniousContextTree(params.getAlphabetContainer(),(byte)Math.min(i,order));
				tree[i].setTreeNumber(i);
				tree[i].setRuntimeImprovements(memoization,pruning);
				tree[i].maximize(dataForThisTree,weights);
				if(progress!=null)
					progress.setCurrent(i);
			}
		trained = true;
		
	}
	
	/**
	 * The maximal/initial order of the iPMM.
	 * @return order
	 */
	public byte getOrder() {
		return (Byte)params.getParameterAt(0).getValue();
	}

	public double getLogScoreOfCurrentStructure() {
		double res = 0;
		for(int i=0;i<tree.length;i++) {
			res+=tree[i].getScore();
//			res-=Math.log(1+Math.min(i, order));
		}
//		System.out.println(res);
		return res;
	}

	public InstanceParameterSet<? extends InstantiableFromParameterSet> getCurrentParameterSet() throws Exception {
		return params.clone();
	}

	public String toString(NumberFormat nf) {
		return getSparseParameterRepresentation();
	}
	
	/**
	 * Computes the position-specific marginal distributions coinciding with this iPMM.
	 * @return double array where the first index is the sequence position and the second index is the nucleotide
	 */
	public double[][] getPositionSpecificMononucleotideMarginals() {
		double[][] res = new double[length][];
		for(int i=0;i<res.length;i++) {
			int locL = Math.min(i, order);
			SequenceIterator it = new SequenceIterator(locL);
			res[i]= new double[(int)alphabets.getAlphabetAt(i).length()];
			if(i==0||order==0) {
				for(int a=0;a<res[i].length;a++) {
					try {
						Sequence seq = Sequence.create(alphabets, alphabets.getSymbol(i,a));
						res[i][a]=Math.exp(tree[i].getLogProbFor(seq, 0));
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			} else {
				int[] bounds = new int[locL];
				for(int j=0;j<bounds.length;j++) {
					bounds[j]=(int)alphabets.getAlphabetAt(i).length();
				}
				it.setBounds(bounds);
				for(int a=0;a<res[i].length;a++) {
					it.reset();
					res[i][a]=0;
					do {
						double condFactor = 1;
						for(int k=0;k<bounds.length;k++) {
							condFactor*=res[i-bounds.length+k][it.discreteValAt(k)];
						}
						String str = "";
						for(int j=0;j<bounds.length;j++) {
							str+= alphabets.getSymbol(j, it.discreteValAt(j));
						} 
						str+=alphabets.getSymbol(i,a);
						Sequence seq = null;
						try {
							seq = Sequence.create(alphabets, str);
						} catch(Exception e) {
							e.printStackTrace();
						}
						try {
							res[i][a]+=condFactor*Math.exp(tree[i].getLogProbFor(seq, seq.getLength()-1));
							} catch(Exception e) {
								e.printStackTrace();
						}
					} while(it.next());
				}
				Normalisation.sumNormalisation(res[i]);
			}
		}
		return res;
	}
	
	/**
	 * Enables or disables algorithmic techniques for speeding up PCT optimization
	 * @param memoization if <code>true</code>, memoization (Eggeling et al. ICML 2015) is enabled
	 * @param pruning if <code>true</code>, pruning (Eggeling et al. UAI 2016) is enabled
	 */
	public void setRuntimeImprovements(boolean memoization, boolean pruning) {
		this.memoization=memoization;
		this.pruning=pruning;
		
	}
	
	/**
	 * Allows to set a {@link ProgressUpdater}.
	 * @param progress an instance of  {@link  ProgressUpdater}
	 */
	public void setProgressUpdater(ProgressUpdater progress) {
		this.progress=progress;
	}

}
