/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.tools;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Date;
import java.util.LinkedList;

import de.jstacs.DataType;
import de.jstacs.classifiers.trainSMBased.TrainSMBasedClassifier;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.io.StringExtractor;
import de.jstacs.parameters.FileParameter;
import de.jstacs.parameters.Parameter;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.ParameterSetContainer;
import de.jstacs.parameters.SelectionParameter;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.SimpleParameterSet;
import de.jstacs.parameters.validation.NumberValidator;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.projects.inmode.utils.DataReader;
import de.jstacs.results.Result;
import de.jstacs.results.ResultSet;
import de.jstacs.results.TextResult;
import de.jstacs.sequenceScores.statisticalModels.trainable.AbstractTrainableStatisticalModel;
import de.jstacs.sequenceScores.statisticalModels.trainable.PFMWrapperTrainSM;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.HomogeneousMM;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.parameters.HomMMParameterSet;
import de.jstacs.tools.JstacsTool;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.tools.ToolResult;
/**
 * InMoDe tool ClassicationApp
 * @author Ralf Eggeling
 *
 */
public class ClassificationApp implements JstacsTool {

	public ParameterSet getToolParameters() {
		
		
		LinkedList<Parameter> parameters = new LinkedList<Parameter>();
		try {
			parameters.add(new FileParameter("Input data", "The input data file containing the sequences to be classified.", "*", true));
			parameters.add(new FileParameter("Foreground model", "The foreground model file.", "xml", true));
			SimpleParameterSet p0 = new SimpleParameterSet();
			SimpleParameterSet p1 = new SimpleParameterSet(new FileParameter("Model", "The model file.", "xml", true));
			SimpleParameterSet p2 = new SimpleParameterSet(new FileParameter("Data", "The data file for learning a background model from.", "*", true),new SimpleParameter(DataType.BYTE, "Order","The order of the homogeneous background model.",true,new NumberValidator<Byte>((byte)0,(byte)5),(byte)1));
			SelectionParameter sel = new SelectionParameter(DataType.PARAMETERSET, new String[]{"Uniform PWM","Previously learned","Generated from data"}, new ParameterSet[]{p0,p1,p2}, "Background","The type of background model that is used.",true);
			parameters.add(sel);
			parameters.add(new SimpleParameter(DataType.STRING,"Name","An arbitrary name that can make the output identifiable.",false,null,null));
		}  catch(Exception e){
			e.printStackTrace();
			throw new RuntimeException();
		}
		return new SimpleParameterSet(parameters.toArray(new Parameter[0]));
	}

	
	public ToolResult run(ParameterSet parameters, Protocol protocol, ProgressUpdater progress, int threads)
			throws Exception {
		String outputName = (String) parameters.getParameterForName("Name").getValue();
		if(outputName==null||outputName.equals("")) {
			outputName = "Classification";
		}
		protocol.append("Reading data... ");
		AlphabetContainer con = new AlphabetContainer(DNAAlphabet.SINGLETON);
		File f = new File((String) parameters.getParameterForName("Input data").getValue());
		char ignore = StringExtractor.USUALLY;
		if(FileManager.readFile(f).toString().charAt(0)=='>') {
			ignore=StringExtractor.FASTA;
		}
		DataSet data;
		try {
			data = new DataSet(con,new SparseStringExtractor(f,ignore));
		} catch (WrongAlphabetException wae) {
			protocol.append("replace IUPAC symbols... ");
			data =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( f, ignore) );
		}
		StringBuffer sbIn = new StringBuffer();
		sbIn.append(((((FileParameter)parameters.getParameterForName("Foreground model")).getFileContents()).getContent()));
		InhomogeneousPMM fgModel = new InhomogeneousPMM(sbIn);
		AbstractTrainableStatisticalModel bgModel = null;
		SelectionParameter sel = (SelectionParameter)parameters.getParameterForName("Background");

		
		if(sel.getSelected()==0) {
			double[][] pwm = new double[fgModel.getLength()][4];
			for(int i=0;i<pwm.length;i++) {
				Arrays.fill(pwm[i],1); 
			}
			bgModel = new PFMWrapperTrainSM(new AlphabetContainer(DNAAlphabet.SINGLETON),"",pwm,0);
			((PFMWrapperTrainSM)bgModel).getPWM();
		}
		if(sel.getSelected()==1) {
			sbIn = new StringBuffer();
			sbIn.append(((((FileParameter)((SimpleParameterSet)sel.getValue()).getParameterForName("Model")).getFileContents()).getContent()));
			bgModel = new InhomogeneousPMM(sbIn);
		}
		if(sel.getSelected()==2) {
			bgModel = new HomogeneousMM(new HomMMParameterSet(con,1,"",(byte)((SimpleParameterSet)sel.getValue()).getParameterForName("Order").getValue()));
			File fb = new File((String) ((SimpleParameterSet)sel.getValue()).getParameterForName("Data").getValue());
			char ignorebg = StringExtractor.USUALLY;
			if(FileManager.readFile(fb).toString().charAt(0)=='>') {
				ignorebg=StringExtractor.FASTA;
			}
			DataSet bgData;
			try {
				bgData = new DataSet(con,new SparseStringExtractor(fb,ignorebg));
			} catch (WrongAlphabetException wae) {
				protocol.append("replace IUPAC symbols... ");
				bgData =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( fb, ignorebg) );
			}				
			bgModel.train(bgData);
		}
		TrainSMBasedClassifier classifier = new TrainSMBasedClassifier(fgModel,bgModel);
		protocol.append("completed.\n");
		protocol.append("Classify sequences... ");
		byte[] classResults = classifier.classify(data);
		StringBuffer sb = new StringBuffer();
		sb.append("Sequence ID\tClass\n");
		for(int i=0;i<classResults.length;i++) {
			sb.append(i+"\t"+classResults[i]+"\n");
		}
		protocol.append("completed.\n");
		LinkedList<Result> result = new LinkedList<Result>();
		TextResult tr = new TextResult("Classification result of "+outputName,"All class assignments of "+outputName,new FileParameter.FileRepresentation("",sb.toString()),"txt",null,null,true);
		result.add(tr);
		ToolResult res = new ToolResult("Classification result of "+outputName, "Classification result of "+outputName, null, new ResultSet(result), parameters, getToolName(), new Date(System.currentTimeMillis()));
		return res;
	}

	public String getToolName() {
		return "ClassificationApp";
	}

	public String getToolVersion() {
		return "1.1";
	}

	public String getShortName() {
		return "classify";
	}

	public String getDescription() {
		return "performs a binary classification according to two iPMMs."; 
	}

	public String getHelpText() {
		try {
			return FileManager.readInputStream( ScanApp.class.getClassLoader().getResourceAsStream( "de/jstacs/projects/inmode/tools/helpClassificationApp.txt" ) ).toString();
		} catch ( IOException e ) {
			e.printStackTrace();
			return "";
		}
	}

	public ResultEntry[] getDefaultResultInfos() {
		return null;
	}

}
