/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.tools;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.LinkedList;

import de.jstacs.DataType;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.io.StringExtractor;
import de.jstacs.parameters.FileParameter;
import de.jstacs.parameters.Parameter;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.SimpleParameterSet;
import de.jstacs.parameters.validation.NumberValidator;
import de.jstacs.projects.inmode.models.latentVariables.MultiComponentOOPSModel;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhPMMParameterSet;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.projects.inmode.utils.CSLPlotter;
import de.jstacs.projects.inmode.utils.DataReader;
import de.jstacs.results.PlotGeneratorResult;
import de.jstacs.results.Result;
import de.jstacs.results.ResultSet;
import de.jstacs.results.ResultSetResult;
import de.jstacs.results.StorableResult;
import de.jstacs.results.TextResult;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.HomogeneousMM;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.parameters.HomMMParameterSet;
import de.jstacs.tools.JstacsTool;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.tools.ToolResult;
import de.jstacs.utils.SeqLogoPlotter;
import de.jstacs.utils.graphics.RasterizedAdaptor;
/**
 * InMoDe tool DeNovoMoDe
 * @author Ralf Eggeling
 *
 */
public class DeNovoMoDe implements JstacsTool {

	public ParameterSet getToolParameters() {
		
		LinkedList<Parameter> parameters = new LinkedList<Parameter>();
		try {
			parameters.add(new FileParameter("Input data", "The file containing the input data.", "*", true));
			parameters.add(new SimpleParameter(DataType.INT,"Motif width","The width of the motif.",true,new NumberValidator<Integer>((int)4,100),(int)15));
			parameters.add(new SimpleParameter(DataType.BYTE,"Motif order","The maximal order of iPMM.",true,new NumberValidator<Byte>((byte)0,(byte)8),(byte)2));
			parameters.add(new SimpleParameter(DataType.BYTE,"Flanking order","The order of the model for the flanking sequences.",true,new NumberValidator<Byte>((byte)0,(byte)3),(byte)2));
			parameters.add(new SimpleParameter(DataType.INT,"Initial iterations","The number of initial iteration steps in the stochastic search.",true,new NumberValidator<Integer>(1,Integer.MAX_VALUE),50));
			parameters.add(new SimpleParameter(DataType.INT,"Additional iterations","The number of additional iteration steps that need to pass since the last optimal iteration step before termination.",true,new NumberValidator<Integer>(0,Integer.MAX_VALUE),10));
			parameters.add(new SimpleParameter(DataType.INT,"Number of restarts","The number of restarts of the stochastic search for reducing the risk of finding local optima.",true,new NumberValidator<Integer>(1,Integer.MAX_VALUE),3));
			parameters.add(new SimpleParameter(DataType.STRING,"Name","An arbitrary name that can make the output identifiable.",false,null,null));
		}  catch(Exception e){
			throw new RuntimeException();
		}
		return new SimpleParameterSet(parameters.toArray(new Parameter[0]));
	}

	
	public ToolResult run(ParameterSet parameters, Protocol protocol, ProgressUpdater progress, int threads)
			throws Exception {
		
		String outputName = (String) parameters.getParameterForName("Name").getValue();
		if(outputName==null||outputName.equals("")) {
			outputName = "DeNovo("+parameters.getParameterForName("Motif width").getValue()+","+(byte)parameters.getParameterForName("Motif order").getValue()+","+(byte)parameters.getParameterForName("Flanking order").getValue()+")";
		}
		protocol.append("Reading data... ");
		AlphabetContainer con = new AlphabetContainer(DNAAlphabet.SINGLETON);
		File f = new File((String) parameters.getParameterForName("Input data").getValue());
		char ignore = StringExtractor.USUALLY;
		if(FileManager.readFile(f).toString().charAt(0)=='>') {
			ignore=StringExtractor.FASTA;
		}
		DataSet data;
		try {
			data = new DataSet(con,new SparseStringExtractor(f,ignore));
		} catch (WrongAlphabetException wae) {
			protocol.append("replace IUPAC symbols... ");
			data =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( f, ignore) );
		}
		protocol.append("completed.\n");
		protocol.append("Initializing algorithm... ");
		InhPMMParameterSet params = new InhPMMParameterSet(con,(int)parameters.getParameterForName("Motif width").getValue(),(byte)parameters.getParameterForName("Motif order").getValue());
		InhomogeneousPMM[] motif = new InhomogeneousPMM[]{new InhomogeneousPMM(params)};
		HomogeneousMM flanking = new HomogeneousMM(new HomMMParameterSet(con,1,"Flanking model",(byte)parameters.getParameterForName("Flanking order").getValue()));
		MultiComponentOOPSModel model = new MultiComponentOOPSModel(motif,flanking,true,(int)parameters.getParameterForName("Initial iterations").getValue(),(int)parameters.getParameterForName("Additional iterations").getValue(),(int)parameters.getParameterForName("Number of restarts").getValue());
		model.setProtocol(protocol);
		progress.setLast((int)parameters.getParameterForName("Number of restarts").getValue()*(int)parameters.getParameterForName("Initial iterations").getValue());
		progress.setCurrent(0);
		model.setProgressUpdater(progress);
		protocol.append("completed.\n");
		protocol.append("Learning:\n");
		model.train(data);
		LinkedList<Result> result = new LinkedList<Result>();
		result.add(new TextResult("Logfile of "+outputName+" stochastic search","Logfile for the stochastic search of "+outputName,new FileParameter.FileRepresentation("",model.getLog()),"txt",null,null,true));
		String lvs = model.getBestLVAssignment();
		if(lvs!=null)
			result.add(new TextResult("Latent variables of "+outputName,"Optimal latent variable configuration of "+outputName,new FileParameter.FileRepresentation("",model.getBestLVAssignment()),"txt",null,null,true));
				
		LinkedList<Result> resultLocal = new LinkedList<Result>();
		InhomogeneousPMM local = (InhomogeneousPMM) model.getMotifModels()[0];
		resultLocal.add(new StorableResult("XML of "+outputName+" motif","XML representation of the of "+outputName+", motif",local));
		resultLocal.add(new TextResult("Parameters of "+outputName+" motif","Conditional probability parameters for all PCTs leaves of "+outputName+" motif",new FileParameter.FileRepresentation("",local.getParameterRepresentation()),"txt",null,null,true));
		resultLocal.add(new TextResult("Binding sites of "+outputName+" motif","Binding sites underlying "+outputName+" motif",new FileParameter.FileRepresentation("",model.getBindingSites()[0]),"txt",null,null,true));
		double[][] pwm = local.getPositionSpecificMononucleotideMarginals();
		double[][] pwmrc = new double[pwm.length][pwm[0].length];
		for(int i=0;i<pwm.length;i++) {
			for(int a=0;a<pwm[i].length;a++)
			pwmrc[pwm.length-1-i][pwm[i].length-1-a]=pwm[i][a];
		}
		{
			RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
			SeqLogoPlotter.SeqLogoPlotGenerator slpg = new SeqLogoPlotter.SeqLogoPlotGenerator(pwm,500);
			slpg.generatePlot(ra);
			PlotGeneratorResult pnr = new PlotGeneratorResult("SL of "+outputName+" motif","The sequence logo of the position-specific mononucleotide marginal distribution of "+outputName+" motif",slpg,true);
			resultLocal.add(pnr); 
		}
		{	RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
			SeqLogoPlotter.SeqLogoPlotGenerator slpg = new SeqLogoPlotter.SeqLogoPlotGenerator(pwmrc,500);
			slpg.generatePlot(ra);
			PlotGeneratorResult pnr = new PlotGeneratorResult("SL(rc) of "+outputName,"The reverse complement sequence logo of the position-specific mononucleotide marginal distribution of "+outputName,slpg,true);
			resultLocal.add(pnr); 
		}
		
		if((byte) parameters.getParameterForName("Motif order").getValue()>0) {	
			RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
			CSLPlotter.CSLPlotGenerator slpg = new CSLPlotter.CSLPlotGenerator((InhomogeneousPMM)model.getMotifModels()[0]);
			slpg.generatePlot(ra);
			PlotGeneratorResult pnr = new PlotGeneratorResult("CSL of "+outputName,"Conditional Sequence logo of "+outputName,slpg,true);
			resultLocal.add(pnr);
		}
		ResultSet rs = new ResultSet(resultLocal);
		ResultSetResult rsr = new ResultSetResult("Learned "+outputName+" motif", outputName+" motif",null, rs) ;
		result.add(rsr);
		ToolResult res = new ToolResult("Learned "+outputName, "All results of "+outputName, null, new ResultSet(result), parameters, getToolName(), new Date(System.currentTimeMillis()));
		return res;
	}

	public String getToolName() {
		return "DeNovoMoDe";
	}

	public String getToolVersion() {
		return "1.1";
	}

	public String getShortName() {
		return "denovo";
	}

	public String getDescription() {
		return "infers an iPMM via de novo motif discovery.";
	}

	public String getHelpText() {
		try {
			return FileManager.readInputStream( DeNovoMoDe.class.getClassLoader().getResourceAsStream( "de/jstacs/projects/inmode/tools/helpDeNovoMoDe.txt" ) ).toString();
		} catch ( IOException e ) {
			e.printStackTrace();
			return "";
		}
	}

	@Override
	public ResultEntry[] getDefaultResultInfos() {
		return null;
	}
	
	

}
