/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.tools;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.LinkedList;

import de.jstacs.DataType;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.io.StringExtractor;
import de.jstacs.parameters.FileParameter;
import de.jstacs.parameters.Parameter;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.SimpleParameterSet;
import de.jstacs.parameters.validation.NumberValidator;
import de.jstacs.projects.inmode.models.latentVariables.MultiComponentOOPSModel;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhPMMParameterSet;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.projects.inmode.utils.CSLPlotter;
import de.jstacs.projects.inmode.utils.DataReader;
import de.jstacs.results.PlotGeneratorResult;
import de.jstacs.results.Result;
import de.jstacs.results.ResultSet;
import de.jstacs.results.ResultSetResult;
import de.jstacs.results.StorableResult;
import de.jstacs.results.TextResult;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.HomogeneousMM;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.parameters.HomMMParameterSet;
import de.jstacs.tools.JstacsTool;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.tools.ToolResult;
import de.jstacs.utils.SeqLogoPlotter;
import de.jstacs.utils.graphics.RasterizedAdaptor;
/**
 * InMoDe tool MixtureMoDe
 * @author Ralf Eggeling
 *
 */
public class MixtureMoDe implements JstacsTool {

	public ParameterSet getToolParameters() {
		
		LinkedList<Parameter> parameters = new LinkedList<Parameter>();
		try {
			parameters.add(new FileParameter("Input data", "The file containing the input data.", "*", true));
			parameters.add(new SimpleParameter(DataType.BYTE,"Number of mixture components","The number of possible mixture components.",true,new NumberValidator<Byte>((byte)1,(byte)100),(byte)2));
			parameters.add(new SimpleParameter(DataType.BYTE,"Component order","The maximal order of the iPMM component models.",true,new NumberValidator<Byte>((byte)0,(byte)8),(byte)2));
			parameters.add(new SimpleParameter(DataType.INT,"Initial iterations","The number of initial iteration steps in the stochastic search.",true,new NumberValidator<Integer>(1,Integer.MAX_VALUE),50));
			parameters.add(new SimpleParameter(DataType.INT,"Additional iterations","The number of additional iteration steps that need to pass since the last optimal iteration step before termination.",true,new NumberValidator<Integer>(0,Integer.MAX_VALUE),10));
			parameters.add(new SimpleParameter(DataType.INT,"Number of restarts","The number of restarts of the stochastic search for reducing the risk of finding local optima.",true,new NumberValidator<Integer>(1,Integer.MAX_VALUE),3));
			parameters.add(new SimpleParameter(DataType.STRING,"Name","An arbitrary name that can make the output identifiable.",false,null,null));
		}  catch(Exception e){
			throw new RuntimeException();
		}
		return new SimpleParameterSet(parameters.toArray(new Parameter[0]));
	}

	public ToolResult run(ParameterSet parameters, Protocol protocol, ProgressUpdater progress, int threads)
			throws Exception {
		
		String outputName = (String) parameters.getParameterForName("Name").getValue();
		if(outputName==null||outputName.equals("")) {
			outputName = "Mixture("+parameters.getParameterForName("Number of mixture components").getValue()+","+(byte)parameters.getParameterForName("Component order").getValue()+")";
		}
		
		AlphabetContainer con = new AlphabetContainer(DNAAlphabet.SINGLETON);
		protocol.append("Reading data... ");
		File f = new File((String) parameters.getParameterForName("Input data").getValue());
		char ignore = StringExtractor.USUALLY;
		if(FileManager.readFile(f).toString().charAt(0)=='>') {
			ignore=StringExtractor.FASTA;
		}
		DataSet data;
		try {
			data = new DataSet(con,new SparseStringExtractor(f,ignore));
		} catch (WrongAlphabetException wae) {
			protocol.append("replace IUPAC symbols... \n");
			data =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( f, ignore) );
		}
		if(data.getElementLength()==0) 
			throw new Exception("For this tool, all input sequences need to be of same length.");
		protocol.append("completed.\n");
		protocol.append("Initializing algorithm... ");
		byte K = (byte)parameters.getParameterForName("Number of mixture components").getValue();
		InhomogeneousPMM[] component = new InhomogeneousPMM[K];
		for(int k=0;k<K;k++) {
			component[k]=new InhomogeneousPMM(new InhPMMParameterSet(con,data.getElementLength(),(byte)parameters.getParameterForName("Component order").getValue()));
			component[k].setRuntimeImprovements(false, true);
		}
		HomogeneousMM flanking = new HomogeneousMM(new HomMMParameterSet(con,1,"",(byte)0));	
		MultiComponentOOPSModel model = new MultiComponentOOPSModel(component,flanking,false,(int)parameters.getParameterForName("Initial iterations").getValue(),(int)parameters.getParameterForName("Additional iterations").getValue(),(int)parameters.getParameterForName("Number of restarts").getValue());
		model.updateLVParam(true);
		model.setProtocol(protocol);
		progress.setLast((int)parameters.getParameterForName("Number of restarts").getValue()*(int)parameters.getParameterForName("Initial iterations").getValue());
		progress.setCurrent(0);
		model.setProgressUpdater(progress);
		protocol.append("completed.\n");
		protocol.append("Learning:\n");
		model.train(data);
		LinkedList<Result> result = new LinkedList<Result>();
		result.add(new TextResult("Logfile of "+outputName,"Logfile for the stochastic search of "+outputName,new FileParameter.FileRepresentation("",model.getLog()),"txt",null,null,true));
		String lvs = model.getBestLVAssignment();
		if(lvs!=null)
			result.add(new TextResult("Latent variables of "+outputName,"Optimal latent variable configuration of "+outputName,new FileParameter.FileRepresentation("",model.getBestLVAssignment()),"txt",null,null,true));
		for(int k=0;k<K;k++) {
			if(model.checkComponentActivity(k)) {
				LinkedList<Result> resultLocal = new LinkedList<Result>();
				InhomogeneousPMM local = (InhomogeneousPMM) model.getMotifModels()[k];
				resultLocal.add(new StorableResult("XML of "+outputName+" component "+k,"XML representation of the of "+outputName+", component "+k,local));
				resultLocal.add(new TextResult("Parameters of "+outputName+" component "+k,"Conditional probability parameters for all PCTs leaves of "+outputName+", component "+k,new FileParameter.FileRepresentation("",local.getParameterRepresentation()),"txt",null,null,true));
				resultLocal.add(new TextResult("Binding sites of "+outputName+" component "+k,"Binding sites underlying "+outputName+" component "+k,new FileParameter.FileRepresentation("",model.getBindingSites()[k]),"txt",null,null,true));
				double[][] pwm = local.getPositionSpecificMononucleotideMarginals();
				double[][] pwmrc = new double[pwm.length][pwm[0].length];
				for(int i=0;i<pwm.length;i++) {
					for(int a=0;a<pwm[i].length;a++)
					pwmrc[pwm.length-1-i][pwm[i].length-1-a]=pwm[i][a];
				}
				{
					RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
					SeqLogoPlotter.SeqLogoPlotGenerator slpg = new SeqLogoPlotter.SeqLogoPlotGenerator(pwm,500);
					slpg.generatePlot(ra);
					PlotGeneratorResult pnr = new PlotGeneratorResult("SL of "+outputName+" component "+k,"The sequence logo of the position-specific mononucleotide marginal distribution of "+outputName+", component "+k,slpg,true);
					resultLocal.add(pnr);
				}
				{	RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
					SeqLogoPlotter.SeqLogoPlotGenerator slpg = new SeqLogoPlotter.SeqLogoPlotGenerator(pwmrc,500);
					slpg.generatePlot(ra);
					PlotGeneratorResult pnr = new PlotGeneratorResult("SL(rc) of "+outputName+" component "+k,"The reverse complement sequence logo of the position-specific mononucleotide marginal distribution of "+outputName+", component "+k,slpg,true);
					resultLocal.add(pnr); 
				}
				
				if((byte) parameters.getParameterForName("Component order").getValue()>0) {	
					RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
					CSLPlotter.CSLPlotGenerator slpg = new CSLPlotter.CSLPlotGenerator((InhomogeneousPMM)model.getMotifModels()[k]);
					slpg.generatePlot(ra);
					PlotGeneratorResult pnr = new PlotGeneratorResult("CSL of "+outputName+" component "+k,"Conditional Sequence logo of "+outputName+" component "+k,slpg,true);
					resultLocal.add(pnr); }
					ResultSet rs = new ResultSet(resultLocal);
					ResultSetResult rsr = new ResultSetResult("Learned "+outputName+" component "+k, outputName+" component "+k,null, rs) ;
					result.add(rsr);
				}
			}
		ToolResult res = new ToolResult("Learned "+outputName, "All output of the mixture model", null, new ResultSet(result), parameters, getToolName(), new Date(System.currentTimeMillis()));
		return res;
	}

	public String getToolName() {
		return "MixtureMoDe";
	}

	public String getToolVersion() {
		return "1.1";
	}

	public String getShortName() {
		return "mixture";
	}

	public String getDescription() {
		return "learns a mixture model with iPMM components.";
	}

	public String getHelpText() {
		try {
			return FileManager.readInputStream( MixtureMoDe.class.getClassLoader().getResourceAsStream( "de/jstacs/projects/inmode/tools/helpMixtureMoDe.txt" ) ).toString();
		} catch ( IOException e ) {
			e.printStackTrace();
			return "";
		}
	}
	
	@Override
	public ResultEntry[] getDefaultResultInfos() {
		return null;
	}

}
