/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.tools;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.Random;
import java.util.TreeSet;

import de.jstacs.DataType;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.data.alphabets.DiscreteAlphabet;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.data.sequences.SparseSequence;
import de.jstacs.data.sequences.annotation.SimpleSequenceAnnotationParser;
import de.jstacs.io.AbstractStringExtractor;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.io.StringExtractor;
import de.jstacs.parameters.FileParameter;
import de.jstacs.parameters.Parameter;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.SelectionParameter;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.SimpleParameterSet;
import de.jstacs.parameters.validation.NumberValidator;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.projects.inmode.utils.BasicSamplingAlgorithms;
import de.jstacs.projects.inmode.utils.DataReader;
import de.jstacs.results.Result;
import de.jstacs.results.ResultSet;
import de.jstacs.results.TextResult;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.HomogeneousMM;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.homogeneous.parameters.HomMMParameterSet;
import de.jstacs.tools.JstacsTool;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.tools.ToolResult;
import de.jstacs.utils.Normalisation;
/**
 * InMoDe tool ScanApp
 * @author Ralf Eggeling
 *
 */
public class ScanApp  implements JstacsTool {

	public ParameterSet getToolParameters() {
		
		LinkedList<Parameter> parameters = new LinkedList<Parameter>();
		try {
			parameters.add(new FileParameter("Input model", "The file with the XML representation of a learned iPMM.", "xml", true));
			parameters.add(new FileParameter("Input data", "The input data file in which a sequence scan is to be performed.", "*", true));
			
			SimpleParameterSet p0 = new SimpleParameterSet(new SimpleParameter(DataType.BYTE,"Order","Order of the Markov chain for generating background data.",true,new NumberValidator<Byte>((byte)0,(byte)10),(byte)3),new SimpleParameter(DataType.INT,"Size factor","Size of the generated background data set as a factor of the foreground data set size.",true,new NumberValidator<Integer>(1,Integer.MAX_VALUE),1));
			SimpleParameterSet p1 = new SimpleParameterSet(new FileParameter("Data file", "A negative data file for computing the prediction threshold based on a given FPR.", "*", true));
			SelectionParameter sel = new SelectionParameter(DataType.PARAMETERSET, new String[]{"Generating","From file"}, new ParameterSet[]{p0,p1}, "Background","The type of background that is used.",true);
			parameters.add(sel);
			parameters.add(new SimpleParameter(DataType.DOUBLE,"FPR","False positive rate on background data.",true,new NumberValidator<Double>((double)1E-99,(double)1),(double)1E-4));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Both strands","Whether the input data is scanned on both strands.",false,null,true));
			parameters.add(new SimpleParameter(DataType.STRING,"Name","An arbitrary name that can make the output identifiable.",false,null,null));
		}  catch(Exception e){
			throw new RuntimeException();
		}
		return new SimpleParameterSet(parameters.toArray(new Parameter[0]));
	}

	public ToolResult run(ParameterSet parameters, Protocol protocol,
			ProgressUpdater progress, int threads) throws Exception {
		String outputName = (String) parameters.getParameterForName("Name").getValue();
		if(outputName==null||outputName.equals("")) {
			outputName = "SequenceScan("+parameters.getParameterForName("FPR").getValue()+")";
		}
		double prog = 0;
		progress.setLast(Double.MAX_VALUE);
		progress.setCurrent(prog);
		protocol.append("Reading data... ");
		AlphabetContainer con = new AlphabetContainer(DNAAlphabet.SINGLETON);
		StringBuffer sbIn = new StringBuffer();
		sbIn.append(((((FileParameter)parameters.getParameterForName("Input model")).getFileContents()).getContent()));
		InhomogeneousPMM model = new InhomogeneousPMM(sbIn);
		File f = new File((String) parameters.getParameterForName("Input data").getValue());
		char ignore = StringExtractor.USUALLY;
		if(FileManager.readFile(f).toString().charAt(0)=='>') {
			ignore=StringExtractor.FASTA;
		}
		DataSet data;
		try {
			data = new DataSet(con,new SparseStringExtractor(f,ignore));
		} catch (WrongAlphabetException wae) {
			protocol.append("replace IUPAC symbols... \n");
			data =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( f, ignore) );
		}
	
		protocol.append("completed.\n");
		DataSet bg;
		ArrayList<Double> al = new ArrayList<Double>(); 
		int fac = 1; 
		protocol.append("Computing background likelihood... ");
		
		SelectionParameter sel = (SelectionParameter)parameters.getParameterForName("Background");
		
		if(sel.getSelected()==1) {
			fac=1;
		} else {
			fac = (int)((SimpleParameterSet)sel.getValue()).getParameterForName("Size factor").getValue();
			if((boolean)parameters.getParameterForName("Both strands").getValue())
				progress.setLast(fac*2*data.getNumberOfElements());
			else
				progress.setLast(fac*data.getNumberOfElements());
		}
		for(int fa = 0;fa<fac;fa++) {
			if(sel.getSelected()==1) {
				File fbg = new File((String) ((SimpleParameterSet)sel.getValue()).getParameterForName("Data file").getValue());
				char ignorebg = StringExtractor.USUALLY;
				if(FileManager.readFile(f).toString().charAt(0)=='>') {
					ignorebg=StringExtractor.FASTA;
				}
				try {
					bg = new DataSet(con,new SparseStringExtractor(fbg,ignorebg));
				} catch (WrongAlphabetException wae) {
					protocol.append("replace IUPAC symbols... \n");
					bg =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( fbg, ignorebg) );
				}				
			} else {
				HomogeneousMM bgModel = new HomogeneousMM(new HomMMParameterSet(con,1,"", (byte)((SimpleParameterSet)sel.getValue()).getParameterForName("Order").getValue() ));
				bgModel.train(data);
				int[] lengths = new int[data.getNumberOfElements()];
				for(int i=0;i<data.getNumberOfElements();i++)
						lengths[i]=data.getElementAt(i).getLength();
				 bg = bgModel.emitDataSet(data.getNumberOfElements(), lengths);
			}

			for(int i=0;i<bg.getNumberOfElements();i++) {
				for(int j=0;j<bg.getElementAt(i).getLength()-model.getLength()+1;j++) {
					al.add((double) model.getLogProbFor(bg.getElementAt(i),j,j+model.getLength()-1));	
				}
				prog++;
				progress.setCurrent(prog);
			}
			if((boolean) parameters.getParameterForName("Both strands").getValue()) {
				DataSet bgRC = bg.getReverseComplementaryDataSet();
				for(int i=0;i<bgRC.getNumberOfElements();i++) {
					for(int j=0;j<bgRC.getElementAt(i).getLength()-model.getLength()+1;j++) {
						al.add((double) model.getLogProbFor(bgRC.getElementAt(i),j,j+model.getLength()-1));	
					}
					prog++;
					progress.setCurrent(prog);
				}
				
			}
			
		}
		protocol.append("completed.\n");
		protocol.append("Sorting likelihood values... ");
		al.sort(null);
		protocol.append("completed.\n");
		double cutoff = al.get( (int) (Math.floor((1-(Double)parameters.getParameterForName("FPR").getValue())*al.size())));
		protocol.append("Scan foreground data... ");
		StringBuffer sbLoc = new StringBuffer();
		StringBuffer sbBS = new StringBuffer();
		int nF = 0;
		int nR = 0;
		TreeSet<Integer> ts = new TreeSet<Integer>();
		LinkedList<String> llLoc = new LinkedList<String>();
		LinkedList<String> llBS = new LinkedList<String>();
		for(int i=0;i<data.getNumberOfElements();i++) {
			Sequence seq = data.getElementAt(i);
			for(int j=0;j<(seq.getLength()-model.getLength()+1);j++) {
				double LLF = model.getLogProbFor(seq.getSubSequence(j,model.getLength()));
				if(LLF>cutoff) {
					llLoc.add(i+"\t"+j+"\t"+(j+model.getLength())+"\t+\t"+Math.exp(LLF));
					llBS.add(seq.getSubSequence(j,model.getLength()).toString());
					nF++;
					ts.add(i);
					prog++;
					progress.setCurrent(prog);
				}
				if((boolean) parameters.getParameterForName("Both strands").getValue()) {
					double LLR = model.getLogProbFor(seq.getSubSequence(j,model.getLength()).reverseComplement());
					if(LLR>cutoff) {
						llLoc.add(i+"\t"+j+"\t"+(j+model.getLength())+"\t-\t"+Math.exp(LLR));
						llBS.add(seq.getSubSequence(j,model.getLength()).reverseComplement().toString());
						nR++;
						ts.add(i);
						prog++;
						progress.setCurrent(prog);
					}
					
				}
			}

		}
		for(int i=0;i<(nF+nR);i++) {
			sbLoc.append(llLoc.removeFirst());
			sbBS.append(llBS.removeFirst());
			if(i<(nF+nR)-1) {
				sbLoc.append("\n");
				sbBS.append("\n");
			}
				
		}
		protocol.append("completed.\n");
		protocol.append("# hits = "+(nF+nR)+" ("+nF+" on + strand)\n");
		protocol.append("# sequences with at least one hit = "+ts.size()+" (of "+data.getNumberOfElements()+")\n");
		LinkedList<Result> result = new LinkedList<Result>();
		TextResult tr1 = new TextResult("Motif hits from "+outputName,"All motif hits from "+outputName+" with a likelihood over a threshold corresponding to a given or generated background data set.",new FileParameter.FileRepresentation("",sbLoc.toString()),"BED",null,null,true);
		TextResult tr2 = new TextResult("Binding sites from "+outputName,"Extracted binding sites corresponding to the hits of "+outputName,new FileParameter.FileRepresentation("",sbBS.toString()),"txt",null,null,true);
		result.add(tr1);
		result.add(tr2);
		ToolResult res = new ToolResult("Result of "+outputName, "All results of the scan "+outputName, null, new ResultSet(result), parameters, getToolName(), new Date(System.currentTimeMillis()));
		return res;
	}


	public String getToolName() {
		return "ScanApp";
	}

	public String getToolVersion() {
		return "1.1";
	}

	public String getShortName() {
		return "scan";
	}

	public String getDescription() {
		return "finds occurrences of an iPMM motif in given sequences.";
	}

	public String getHelpText() {
		try {
			return FileManager.readInputStream( ScanApp.class.getClassLoader().getResourceAsStream( "de/jstacs/projects/inmode/tools/helpScanApp.txt" ) ).toString();
		} catch ( IOException e ) {
			e.printStackTrace();
			return "";
		}
	}

	public ResultEntry[] getDefaultResultInfos() {
		return null;
	}
	


}
