/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.tools;
import java.io.File;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.Locale;

import de.jstacs.DataType;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.io.StringExtractor;
import de.jstacs.parameters.FileParameter;
import de.jstacs.parameters.Parameter;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.SimpleParameterSet;
import de.jstacs.parameters.validation.NumberValidator;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhPMMParameterSet;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.projects.inmode.utils.CSLPlotter;
import de.jstacs.projects.inmode.utils.DataReader;
import de.jstacs.results.ListResult;
import de.jstacs.results.PlotGeneratorResult;
import de.jstacs.results.Result;
import de.jstacs.results.ResultSet;
import de.jstacs.results.StorableResult;
import de.jstacs.results.TextResult;
import de.jstacs.tools.JstacsTool;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.tools.ToolResult;
import de.jstacs.tools.JstacsTool.ResultEntry;
import de.jstacs.utils.SeqLogoPlotter;
import de.jstacs.utils.graphics.RasterizedAdaptor;
/**
 * InMoDe tool SimpleMoDe
 * @author Ralf Eggeling
 *
 */
public class SimpleMoDe implements JstacsTool {

	
	
	public ParameterSet getToolParameters() {
		
		LinkedList<Parameter> parameters = new LinkedList<Parameter>();
		try {
			parameters.add(new FileParameter("Input data", "The file containing the input data.", "*", true));
			parameters.add(new SimpleParameter(DataType.BYTE,"Order","The maximal order of the iPMM.",true,new NumberValidator<Byte>((byte)0,(byte)8),(byte)2));
			parameters.add(new SimpleParameter(DataType.STRING,"Name","An arbitrary name that can make the output identifiable.",false,null,null));
		} catch(Exception e){
			throw new RuntimeException();
		}
		return new SimpleParameterSet(parameters.toArray(new Parameter[0]));
	}

	public ToolResult run(ParameterSet parameters, Protocol protocol, ProgressUpdater progress, int threads) throws Exception {
		
		NumberFormat nf = NumberFormat.getInstance(Locale.US);
		nf.setGroupingUsed(false);
		nf.setMaximumFractionDigits(2);
		nf.setMinimumFractionDigits(2);
		
		String outputName = (String) parameters.getParameterForName("Name").getValue();
		if(outputName==null||outputName.equals("")) {
			outputName = "iPMM("+parameters.getParameterForName("Order").getValue()+")";
		}
		AlphabetContainer con = new AlphabetContainer(DNAAlphabet.SINGLETON);
		protocol.append("Reading data... ");
		File f = new File((String) parameters.getParameterForName("Input data").getValue());
		char ignore = StringExtractor.USUALLY;
		if(FileManager.readFile(f).toString().charAt(0)=='>') {
			ignore=StringExtractor.FASTA;
		}
		DataSet data;
		try {
			data = new DataSet(con,new SparseStringExtractor(f,ignore));
		} catch (WrongAlphabetException wae) {
			protocol.append("replace IUPAC symbols... ");
			data =  DataReader.replaceIUPACSymbols(new SparseStringExtractor( f, ignore) );
		}
		if(data.getElementLength()==0) 
			throw new Exception("Input sequences are not of equal length.");
		protocol.append("completed.\n");
		protocol.append("Initializing algorithm... ");
		double[] w = null;
		InhomogeneousPMM model = new InhomogeneousPMM(new InhPMMParameterSet(con,data.getElementLength(),(byte) parameters.getParameterForName("Order").getValue()));
		model.setRuntimeImprovements(false, true);
		progress.setLast(model.getLength());
		progress.setCurrent(0);
		model.setProgressUpdater(progress);
		protocol.append("completed.\n");
		protocol.append("Learning... ");
		model.train(data,w);
		protocol.append("completed.\n");
		protocol.append("Score of learned model: "+nf.format(model.getLogScoreOfCurrentStructure())+"\n");
		LinkedList<Result> result = new LinkedList<Result>();
		result.add(new TextResult("Parameters of "+outputName,"Conditional probability parameters for all PCTs leaves of "+outputName,new FileParameter.FileRepresentation("",model.getParameterRepresentation()),"txt",null,null,true));
		result.add(new StorableResult("XML of "+outputName,"XML representation of "+outputName,model));
		double[][] pwm = model.getPositionSpecificMononucleotideMarginals();
		double[][] pwmrc = new double[pwm.length][pwm[0].length];
		for(int i=0;i<pwm.length;i++) {
			for(int a=0;a<pwm[i].length;a++)
			pwmrc[pwm.length-1-i][pwm[i].length-1-a]=pwm[i][a];
		}
		{	RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
		SeqLogoPlotter.SeqLogoPlotGenerator slpg = new SeqLogoPlotter.SeqLogoPlotGenerator(pwm,500);
		slpg.generatePlot(ra);
		PlotGeneratorResult pnr = new PlotGeneratorResult("SL of "+outputName,"The sequence logo of the position-specific mononucleotide marginal distribution of "+outputName,slpg,true);
		result.add(pnr); 
		}
		{	RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
		SeqLogoPlotter.SeqLogoPlotGenerator slpg = new SeqLogoPlotter.SeqLogoPlotGenerator(pwmrc,500);
		slpg.generatePlot(ra);
		PlotGeneratorResult pnr = new PlotGeneratorResult("SL(rc) of "+outputName,"The reverse complement sequence logo of the position-specific mononucleotide marginal distribution of "+outputName,slpg,true);
		result.add(pnr); 
		}
		if((byte) parameters.getParameterForName("Order").getValue()>0)
		{	RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
		CSLPlotter.CSLPlotGenerator slpg = new CSLPlotter.CSLPlotGenerator(model);
		slpg.generatePlot(ra);
		PlotGeneratorResult pnr = new PlotGeneratorResult("CSL of "+outputName,"Conditional Sequence logo of "+outputName,slpg,true);
		result.add(pnr); } 
		
		ToolResult res = new ToolResult("Learned "+outputName, "All results from learning "+outputName, null, new ResultSet(result), parameters, getToolName(), new Date(System.currentTimeMillis()));
		return res;
	}

	public String getToolName() {
		return "SimpleMoDe";
	}

	public String getToolVersion() {
		return "1.1";
	}

	public String getShortName() {
		return "simple";
	}

	public String getDescription() {
		return "learns a single iPMM from pre-aligned sequences.";
	}

	public String getHelpText() {
		try {
			return FileManager.readInputStream( SimpleMoDe.class.getClassLoader().getResourceAsStream( "de/jstacs/projects/inmode/tools/helpSimpleMoDe.txt" ) ).toString();
		} catch ( IOException e ) {
			e.printStackTrace();
			return "";
		}
		
		
	}
	
	@Override
	public ResultEntry[] getDefaultResultInfos() {
		return null;
	}
	

}
