/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.tools;

import java.io.IOException;
import java.util.Date;
import java.util.LinkedList;

import de.jstacs.DataType;
import de.jstacs.io.FileManager;
import de.jstacs.parameters.FileParameter;
import de.jstacs.parameters.Parameter;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.SelectionParameter;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.parameters.SimpleParameterSet;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.projects.inmode.utils.CSLPlotter;
import de.jstacs.results.PlotGeneratorResult;
import de.jstacs.results.ResultSet;
import de.jstacs.tools.JstacsTool;
import de.jstacs.tools.ProgressUpdater;
import de.jstacs.tools.Protocol;
import de.jstacs.tools.ToolResult;
import de.jstacs.utils.graphics.RasterizedAdaptor;
/**
 * InMoDe tool VisualizationApp
 * @author Ralf Eggeling
 *
 */
public class VisualizationApp implements JstacsTool {

	
	public ParameterSet getToolParameters() {
		LinkedList<Parameter> parameters = new LinkedList<Parameter>();
		try {
			parameters.add(new FileParameter("Input model", "The file with the XML representation of a learned iPMM.", "xml", true));
			SelectionParameter sel = new SelectionParameter(DataType.INT, new String[]{"Standard","Non-glossy","ColorBrewer Set1","Grayscale"}, new Integer[]{0,1,2,3}, "Color scheme","The color scheme for drawing nucleotide stacks.",true);
			parameters.add(sel);
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"IUPAC","Plot node labels in IUPAC code.",true,null,false));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Marginals","Plot traditional PWM-corresponding nucleotide stacks.",true,null,true));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Trivial","Plot trivial PCTs and corresponding nucleotide stacks.",true,null,false));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Pseudonodes","Plot pseudonodes (full nodes with trivial subtree below).",true,null,false));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"IC scaled","Scales the nucleotide stacks according to the information content.",true,null,true));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Y labels","Show detailed descriptions of plot elements on the Y-axis.",true,null,false));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Grid","Plot grid.",true,null,true));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Positions","Plot sequence position numbers.",true,null,true));
			parameters.add(new SimpleParameter(DataType.BOOLEAN,"Context labels","Plot context labels.",true,null,true));
			parameters.add(new SimpleParameter(DataType.STRING,"Name","An arbitrary name that can make the output identifiable.",false,null,null));
		} catch (Exception e) {
			throw new RuntimeException();
		}
		return new SimpleParameterSet(parameters.toArray(new Parameter[0]));
	}

	public ToolResult run(ParameterSet parameters, Protocol protocol, ProgressUpdater progress, int threads)
			throws Exception {
		String outputName = (String) parameters.getParameterForName("Name").getValue();
		if(outputName==null||outputName.equals("")) {
			outputName = "CSL";
		}
		StringBuffer sbIn = new StringBuffer();
		sbIn.append(((((FileParameter)parameters.getParameterForName("Input model")).getFileContents()).getContent()));
		InhomogeneousPMM model = new InhomogeneousPMM(sbIn);
		RasterizedAdaptor ra = new RasterizedAdaptor(("png"));
		CSLPlotter.CSLPlotGenerator slpg = new CSLPlotter.CSLPlotGenerator(
				model,
				(boolean)parameters.getParameterForName("Grid").getValue(),
				(boolean)parameters.getParameterForName("Y labels").getValue(),
				(boolean)parameters.getParameterForName("Positions").getValue(),
				(boolean)parameters.getParameterForName("Context labels").getValue(),
				(boolean)parameters.getParameterForName("Marginals").getValue(),
				(boolean)parameters.getParameterForName("Trivial").getValue(),
				(boolean)parameters.getParameterForName("IUPAC").getValue(),
				(boolean)parameters.getParameterForName("Pseudonodes").getValue(),
				(boolean)parameters.getParameterForName("IC scaled").getValue(),
				(int)parameters.getParameterForName("Color scheme").getValue());
		slpg.generatePlot(ra);
		PlotGeneratorResult pnr = new PlotGeneratorResult(outputName,outputName,slpg,true);
		ToolResult res = new ToolResult("Result of "+outputName, "All results of the scan "+outputName, null, new ResultSet(pnr), parameters, getToolName(), new Date(System.currentTimeMillis())); 
		return  res;
	}

	public String getToolName() {
		return "VisualizationApp";
	}

	
	public String getToolVersion() {
		return "1.1";
	}

	public String getShortName() {
		return "visualize";
	}

	public String getDescription() {
		return "plots a conditional sequence logo with user-specified appearance.";
	}

	public String getHelpText() {
		try {
			return FileManager.readInputStream( VisualizationApp.class.getClassLoader().getResourceAsStream( "de/jstacs/projects/inmode/tools/helpVisualizationApp.txt" ) ).toString();
		} catch ( IOException e ) {
			e.printStackTrace();
			return "";
		}
	}

	public ResultEntry[] getDefaultResultInfos() {
		return null;
	}

}
