/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.utils;
import java.util.Random;

/**
 * This class provides several algorithms for sampling discrete random variables from given distributions.
 * All methods are static and public, so they can be used in arbitrary contexts. 
 * @author Ralf Eggeling
 *
 */


public class BasicSamplingAlgorithms {
	
	/**
	 * Samples a random variable according to a given probability distribution. 
	 * @param prob Probability distribution that is to be sampled from 
	 * @param rand For generating random numbers
	 * @return i with probability prob[i]
	 */
	public static int sampleDiscreteRandomVariable(double[] prob, Random rand) {
		double r = rand.nextDouble();
		double cumulativeProb = 0;
		int res = 0;
		while(cumulativeProb<r) {
			cumulativeProb+=prob[res];
			res++; 
		};
		res--;
		return res;
	}
	
	/**
	 * Samples the value of a random variable according to a given log probability distribution. 
	 * @param logProb prob Probability distribution that is to be sampled from 
	 * @param rand For generating random numbers
	 * @return i with log probability logProb[i]
	 */
	public static int sampleDiscreteRandomVariableFromLogProbs(double[] logProb, Random rand) {
		double r = rand.nextDouble();
		double cumulativeProb = 0;
		int res = 0;
		while(cumulativeProb<r) {
			cumulativeProb+=Math.exp(logProb[res]);
			res++; 
		};
		res--;
		return res;
	}
	
	/**
	 * Samples a random variable with K possible values from a uniform distribution.
	 * @param K number of possible values
	 * @param rand Instance of {@link Random}
	 * @return A sample from (1,...,K)
	 */
	public static int sampleDiscreteRandomVariableFromUniformDistribution(int K, Random rand) {
		double r = rand.nextDouble();
		return (int)Math.round(r*K-0.5);
	}
}
