/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.utils;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.util.Arrays;
import java.util.TreeSet;

import de.jstacs.io.NonParsableException;
import de.jstacs.io.XMLParser;
import de.jstacs.projects.inmode.models.latentVariables.MultiComponentOOPSModel;
import de.jstacs.projects.inmode.models.variableStructure.parsimonious.inhomogeneous.InhomogeneousPMM;
import de.jstacs.results.PlotGeneratorResult.PlotGenerator;
import de.jstacs.utils.graphics.GraphicsAdaptor;

/**
 * Plots a Conditional Sequence Logo (CSL)
 * @author Ralf Eggeling
 *
 */
public class CSLPlotter {

	private static boolean PLOTGRID;
	
	private static boolean PLOTYLABELS;
	
	private static boolean PLOTPOSITIONS;
	
	private static boolean PLOTCONTEXTLABELS;
	
	private static boolean PLOTMARGINALS;
	
	private static boolean PLOTTRIVIALPCTS;
	
	private static boolean PLOTIUPAC;
	
	private static boolean PLOTPSEUDONODES;
	
	private static boolean PLOTSTACKSCALED;
	
	private static int PLOTWIDTH;
	
	private static int PLOTHEIGHT;
	
	private static int NODELABELFONTSIZE = 16;
	
	private static int LABELFONTSIZE = 16;
	
	private static int PCTDEPTH = 2;
	
	private static int MAXPCTDEPTH = 2;
	
	private static int NUMBEROFLEAVES = 4;
	
	private static int STACKHEIGHT = 140;
	
	private static int MARGINALSTACKHEIGHT = 140;
	
	private static int STACKWIDTH = 60;
	
	private static int STACKBORDER = 5;
	
	private static int LEFTLEGENDSPACE = 0;
	
	private static int TOPLEGENDSPACE = 40;
	
	private static int SPACEBETWEEN = 10;
	
	private static int LAYERDISTANCE = 20;
	
	private static int COLORSCHEME = 0;
	
	private static float STROKE = (float) 1.5;
	
	private static int NODEHEIGHT = NODELABELFONTSIZE+10;
	
	private static double[][] PROB;
	
	private static String[] CONTEXTS;

	private static int[][] LEAVESBELOW;
	
	private static String[][] NODELABELS;
	
	private static String[][] PARENTS;
	
	private static int TOTALNUMBEROFLEAVES = 1;
	
	private static double[][] PWMMARGINALS;
	
	private static int POS = 0;
	
	private static String[] ALLSTRUCTPARAMS;
	
	private static int[] numbers = {  100,   90,  50,   40,   10,    9,    5,    4,    1 };

    private static String[] letters = { "C",  "XC", "L",  "XL",  "X",  "IX", "V",  "IV", "I" };
	
	public static class CSLPlotGenerator implements PlotGenerator {

		String[] allStructParams;
		double[][] pwmMarginals;
		byte maxDepth;
		int totalNumberOfLeaves;
		boolean plotgrid=true;
		boolean plotylabels=false;
		boolean plotpositions=true;
		boolean plotcontextlabels=true;
		boolean plotmarginals=true;
		boolean plotiupac=false;
		boolean plottrivialpcts=false;
		boolean plotpseudonodes=false;
		boolean plotstackscaled=true;
		int usedcolorscheme=0;
		
		
		/**
		 * The standard constructor for the interface {@link de.jstacs.Storable}. Creates a new {@link  CSLPlotGenerator} out of a {@link StringBuffer}.
		 * @param sb the {@link StringBuffer} to be parsed
		 * @throws NonParsableException  if the {@link  CSLPlotGenerator} could not be reconstructed out of
		 *             the {@link StringBuffer} <code>sb</code>      
		 */		
		public CSLPlotGenerator(StringBuffer sb) throws NonParsableException {
			this.allStructParams = (String[]) XMLParser.extractObjectForTags(sb, "allStructParams");
		}
		
		public StringBuffer toXML() {
			StringBuffer xml = new StringBuffer();
			XMLParser.appendObjectWithTags(xml, allStructParams, "allStructParams");
			return xml;
		}
		
		/**
		 * Standard constructor for a {@link CSLPlotGenerator} with standard plot parameters.
		 * @param model The {@link InhomogeneousPMM} from which the CSL is to be plotted
		 */
		public CSLPlotGenerator(InhomogeneousPMM model) {
			totalNumberOfLeaves = model.getNumberOfLeaves();
			allStructParams = model.getSparseParameterRepresentation().split("&");
			pwmMarginals = model.getPositionSpecificMononucleotideMarginals();
			maxDepth = model.getOrder();
		
		}
		
		/**
		 * Customized constructor for a {@link CSLPlotGenerator} with plot parameters to be set.
		 * @param model The {@link InhomogeneousPMM} from which the CSL is to be plotted
		 * @param grid if <code>true</code>, grid is plotted
		 * @param ylabels if <code>true</code>, labels describing plot elements are shown
		 * @param positions if <code>true</code>, whether sequence positions are shown
		 * @param contextlabels if <code>true</code>, context labels are shown
		 * @param marginals if <code>true</code>, marginal distributions are shown
		 * @param trivialpcts if <code>true</code>, trivial PCTs are shown
		 * @param iupac if <code>true</code>, node labels are shown in IUPAC code
		 * @param pseudonodes if <code>true</code>, pseudo-nodes are shown
		 * @param stackscaled if <code>true</code>, the nucleotide stacks are scaled according to information content
		 * @param colorscheme the desired color scheme for the nucleotide stacks
		 */
		 public CSLPlotGenerator(InhomogeneousPMM model, boolean grid, boolean ylabels, boolean positions,boolean contextlabels, boolean marginals,  boolean trivialpcts, boolean iupac, boolean pseudonodes, boolean stackscaled, int colorscheme) {
			totalNumberOfLeaves = model.getNumberOfLeaves();
			allStructParams = model.getSparseParameterRepresentation().split("&");
			if(plotmarginals)
				pwmMarginals = model.getPositionSpecificMononucleotideMarginals();
			maxDepth = model.getOrder();
			plotylabels=ylabels;
			plotgrid=grid;
			plotpositions=positions;
			plotmarginals=marginals;
			plotiupac=iupac;
			plotcontextlabels=contextlabels;
			plottrivialpcts=trivialpcts;
			plotpseudonodes=pseudonodes;
			plotstackscaled=stackscaled;
			usedcolorscheme=colorscheme;
			
		 }
		
		private void initialComputations() {
			TOTALNUMBEROFLEAVES = this.totalNumberOfLeaves;
			ALLSTRUCTPARAMS = this.allStructParams;
			MAXPCTDEPTH = this.maxDepth;
			PLOTGRID=plotgrid;					
			PLOTYLABELS=plotylabels;
			if(PLOTYLABELS) 
				LEFTLEGENDSPACE=170;
			else
				LEFTLEGENDSPACE=0;
			PLOTPOSITIONS=plotpositions;
			if(PLOTPOSITIONS) 
				TOPLEGENDSPACE=40;
			else
				TOPLEGENDSPACE=10;
			PLOTCONTEXTLABELS = plotcontextlabels;
			PLOTMARGINALS=plotmarginals;
			if(PLOTMARGINALS) {
				PWMMARGINALS = this.pwmMarginals;
				MARGINALSTACKHEIGHT=STACKHEIGHT;
			} else {
				MARGINALSTACKHEIGHT=0;
			}
			PLOTWIDTH = LEFTLEGENDSPACE+TOTALNUMBEROFLEAVES*STACKWIDTH+(ALLSTRUCTPARAMS.length-1)*SPACEBETWEEN;
			PLOTHEIGHT = TOPLEGENDSPACE+MARGINALSTACKHEIGHT+(NODEHEIGHT+LAYERDISTANCE)*(MAXPCTDEPTH+1)+STACKHEIGHT;
			if(!PLOTMARGINALS)
				PLOTHEIGHT+=LAYERDISTANCE;
			PLOTTRIVIALPCTS=plottrivialpcts;
			PLOTIUPAC=plotiupac;
			PLOTPSEUDONODES=plotpseudonodes;
			PLOTSTACKSCALED=plotstackscaled;
			COLORSCHEME=usedcolorscheme;
		}
			

		
		public void generatePlot(GraphicsAdaptor ga) throws Exception {
			initialComputations();
			plotLogo(ga.getGraphics(PLOTWIDTH, PLOTHEIGHT),PLOTWIDTH,PLOTHEIGHT);			
		}
	}
	
	private static void initialComputations(int pos) {
		POS = pos;
		CONTEXTS = ALLSTRUCTPARAMS[pos].split("%");
		PROB = new double[CONTEXTS.length][];
		NUMBEROFLEAVES = CONTEXTS.length;
		for(int c=0;c<CONTEXTS.length;c++) {
			String[] contextElements = CONTEXTS[c].split("\t");
			PROB[c] = new double[contextElements.length-1];
			for(int i=1;i<contextElements.length;i++) {
				PROB[c][i-1]=Double.parseDouble(contextElements[i]);
			}
			if(contextElements[0].length()>0) {
				CONTEXTS[c]=contextElements[0].substring(1, contextElements[0].length()-1);
				CONTEXTS[c]=CONTEXTS[c].replace("][", " ");
			} else {
				CONTEXTS[c]="";
			}
		}
		PCTDEPTH = CONTEXTS[0].split(" ").length;
		String[][] labels = new String[NUMBEROFLEAVES][];
		for(int c=0;c<NUMBEROFLEAVES;c++) {
			labels[c]=CONTEXTS[c].split(" ");
		}
		NODELABELS = new String[PCTDEPTH][];
		PARENTS = new String[PCTDEPTH][];
		for(int d=0;d<PCTDEPTH;d++) {
			TreeSet<String> seenStrings = new TreeSet<String>();
			TreeSet<String> seenParentStrings = new TreeSet<String>();
			String existingNodes = "";
			String parentIndices = "";
			int currentParentIndex = -1;
			for(int c=0;c<NUMBEROFLEAVES;c++) {
				String current = "";
				String currentP = "";
				for(int f=d;f<PCTDEPTH;f++) {
					current+=labels[c][f]+" ";
				}
				for(int f=d+1;f<PCTDEPTH;f++) {
					currentP+=labels[c][f]+" ";
				}
				boolean notSeen = seenStrings.add(current);
				boolean notSeenP = seenParentStrings.add(currentP);
				if(notSeen) {
					existingNodes+=labels[c][d]+" ";
				}
				if(notSeenP) {
					currentParentIndex++;
				}
				if(notSeen)
					parentIndices+=currentParentIndex+" ";
			}
			NODELABELS[d]=existingNodes.split(" ");
			PARENTS[d]=parentIndices.split(" ");
		}

		LEAVESBELOW = new int[NODELABELS.length][];
		for(int d=0;d<LEAVESBELOW.length;d++) {
			LEAVESBELOW[d] = new int[NODELABELS[d].length];
		}
		Arrays.fill(LEAVESBELOW[0], 1);
		for(int d=0;d<LEAVESBELOW.length-1;d++) {
			for(int c=0;c<NODELABELS[d].length;c++) {
				LEAVESBELOW[d+1][Integer.parseInt(PARENTS[d][c])]+=LEAVESBELOW[d][c];
			}
		}
	}
	
	
	protected static void plotLogo(Graphics2D g, int width, int height) {
		g = (Graphics2D)g.create();		
		g.setColor( Color.WHITE );
		g.fillRect( 0,0 , width, height );
		g.setColor( Color.BLACK );
		g.setStroke( new BasicStroke( STROKE ) );
		g.setFont(new Font(g.getFont().getName(),Font.PLAIN,LABELFONTSIZE));
		int takenXspace = 0;
		if(PLOTYLABELS) {
			if(PLOTPOSITIONS)
				g.drawString("Sequence position j:", 5, 10+LABELFONTSIZE);
			g.setFont(new Font(g.getFont().getName(),Font.PLAIN,NODELABELFONTSIZE));
			if(PLOTMARGINALS) {
				g.drawString("Marginal", 5, (int) (TOPLEGENDSPACE+STACKHEIGHT/2-LABELFONTSIZE*3));
				g.drawString("nucleotide", 5, (int) (TOPLEGENDSPACE+STACKHEIGHT/2-LABELFONTSIZE*1.5));
				g.drawString("probabilities", 5, (int) (TOPLEGENDSPACE+STACKHEIGHT/2+LABELFONTSIZE*0));
				g.drawString("at position j:", 5,(int) (TOPLEGENDSPACE+STACKHEIGHT/2+LABELFONTSIZE*1.5));
			}
			
			g.drawString("Conditional", 5, PLOTHEIGHT-STACKHEIGHT/2-LABELFONTSIZE*3);
			g.drawString("nucleotide", 5,(int)(PLOTHEIGHT-STACKHEIGHT/2-LABELFONTSIZE*1.5));
			g.drawString("probabilities", 5, PLOTHEIGHT-STACKHEIGHT/2);
			g.drawString("at position j:", 5, (int)(PLOTHEIGHT-STACKHEIGHT/2+LABELFONTSIZE*1.5));
			
			for(int d=1;d<=MAXPCTDEPTH;d++) {
				if(PLOTMARGINALS)
					g.drawString("Nucleotide(s) at position j-"+d+":", 5, TOPLEGENDSPACE+MARGINALSTACKHEIGHT+LAYERDISTANCE/2+d*(LAYERDISTANCE+NODEHEIGHT)+NODELABELFONTSIZE/2);
				else 
					g.drawString("Nucleotide(s) at position j-"+d+":", 5, TOPLEGENDSPACE+3*LAYERDISTANCE/2+d*(LAYERDISTANCE+NODEHEIGHT)+NODELABELFONTSIZE/2);
			}
			if(PLOTSTACKSCALED)	{
				if(PLOTMARGINALS)
					CustomizedSeqLogoPlotter.plotYAxis(g, LEFTLEGENDSPACE-50, STACKHEIGHT+TOPLEGENDSPACE, STACKHEIGHT);
				CustomizedSeqLogoPlotter.plotYAxis(g, LEFTLEGENDSPACE-50, PLOTHEIGHT, STACKHEIGHT);
			}
			takenXspace = LEFTLEGENDSPACE;
		}
		
		for(int i=0;i<ALLSTRUCTPARAMS.length;i++) {
			initialComputations(i);
			plotSingleCSL(g,takenXspace,0);
			takenXspace+=STACKWIDTH*NUMBEROFLEAVES+SPACEBETWEEN;
		}
		
		if(PLOTGRID) {
			g.setColor( Color.lightGray );
			Stroke dashed = new BasicStroke(STROKE, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0, new float[]{9}, 0);
			g.setStroke(dashed);
			if(PLOTMARGINALS) {
				g.drawLine( LEFTLEGENDSPACE,TOPLEGENDSPACE ,PLOTWIDTH, TOPLEGENDSPACE);
				g.drawLine( LEFTLEGENDSPACE, TOPLEGENDSPACE+MARGINALSTACKHEIGHT-25 ,PLOTWIDTH, TOPLEGENDSPACE+MARGINALSTACKHEIGHT-25 );
			} else {
				g.drawLine( LEFTLEGENDSPACE,TOPLEGENDSPACE ,PLOTWIDTH, TOPLEGENDSPACE);
			}
	    	
	    	g.drawLine( LEFTLEGENDSPACE, PLOTHEIGHT-25 ,PLOTWIDTH, PLOTHEIGHT-25 );
	    	g.drawLine( LEFTLEGENDSPACE, PLOTHEIGHT-STACKHEIGHT ,PLOTWIDTH, PLOTHEIGHT-STACKHEIGHT );
			g.setColor( Color.BLACK );
			g.setStroke( new BasicStroke( STROKE ) );
		}
	}
	
	private static void plotSingleCSL(Graphics2D g, int x, int y) {
		int usedHorizontalSpace=0;
		g.setColor( Color.BLACK );
		if(PLOTPOSITIONS) {
			g.setFont(new Font(g.getFont().getName(),Font.PLAIN,18));
			String label = (POS+1)+"";
			g.drawString(label, x+(STACKWIDTH*(NUMBEROFLEAVES))/2-(int)(g.getFontMetrics().getStringBounds( label, g ).getWidth()/2), y+10+LABELFONTSIZE);
			g.setFont(new Font(g.getFont().getName(),Font.PLAIN,LABELFONTSIZE));
		}
		if(PLOTMARGINALS)
			CustomizedSeqLogoPlotter.plotStack(g,x+(STACKWIDTH*(NUMBEROFLEAVES-1))/2, y+STACKHEIGHT+TOPLEGENDSPACE, STACKWIDTH, MARGINALSTACKHEIGHT, PWMMARGINALS[POS],"",PLOTSTACKSCALED,STACKBORDER,COLORSCHEME);
		
		int s = 1;
		if(PLOTTRIVIALPCTS)
			s=0;
		if(CONTEXTS.length>s) {
			for(int c=0;c<CONTEXTS.length;c++) {
				g.setFont(new Font(g.getFont().getName(),Font.PLAIN,18));
				String label="";
				if(PLOTCONTEXTLABELS)
					label=getRoman(c+1);
				CustomizedSeqLogoPlotter.plotStack(g,x+usedHorizontalSpace,y+PLOTHEIGHT, STACKWIDTH, STACKHEIGHT, PROB[c],label,PLOTSTACKSCALED,STACKBORDER,COLORSCHEME);
				g.setFont(new Font(g.getFont().getName(),Font.PLAIN,LABELFONTSIZE));
				usedHorizontalSpace+=STACKWIDTH;
			}
			if(PLOTMARGINALS)
				plotPCT(g,x,y+MARGINALSTACKHEIGHT+TOPLEGENDSPACE-LAYERDISTANCE/2);
			else
				plotPCT(g,x,y+TOPLEGENDSPACE+LAYERDISTANCE/2);
		}
		if(PLOTGRID) {
			if(POS>0) {
				g.setColor( Color.lightGray );
				Stroke dashed = new BasicStroke(STROKE, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0, new float[]{9}, 0);
				g.setStroke(dashed);
		    	g.drawLine( x-SPACEBETWEEN/2, 0 ,  x-SPACEBETWEEN/2, PLOTWIDTH );
				g.setColor( Color.BLACK );
				g.setStroke( new BasicStroke( STROKE ) );
			}
		}
	}
	
	private static void plotPCT(Graphics2D g, int x, int y) {
		int[][] xOfNode = new int[PCTDEPTH][];
		for(int d=0;d<xOfNode.length;d++) {
			xOfNode[d] = new int[LEAVESBELOW[d].length];
			int usedSpace = 0;
			for(int f=0;f<xOfNode[d].length;f++) {
				xOfNode[d][f]=x+usedSpace+STACKWIDTH*LEAVESBELOW[d][f]/2;
				usedSpace+=STACKWIDTH*LEAVESBELOW[d][f];
			}
		}
		for(int d=0;d<xOfNode.length;d++) {
			for(int f=0;f<xOfNode[d].length;f++) {
				if(!NODELABELS[d][f].equals("ACGT")||PLOTPSEUDONODES||LEAVESBELOW[d][f]>1) {
					if(d<xOfNode.length-1)
						g.drawLine( xOfNode[d][f], y+(PCTDEPTH-d+1)*LAYERDISTANCE+(PCTDEPTH-d)*NODEHEIGHT ,  xOfNode[d+1][Integer.parseInt(PARENTS[d][f])], y+(PCTDEPTH-d)*LAYERDISTANCE+(PCTDEPTH-d-1)*NODEHEIGHT );
					else
						g.drawLine( xOfNode[d][f], y+(PCTDEPTH-d+1)*LAYERDISTANCE+(PCTDEPTH-d)*NODEHEIGHT ,  x+NUMBEROFLEAVES*STACKWIDTH/2,  y+LAYERDISTANCE );

				}
			}
		}
		for(int d=0;d<xOfNode.length;d++) {
			for(int f=0;f<xOfNode[d].length;f++) {
				if(!NODELABELS[d][f].equals("ACGT")||PLOTPSEUDONODES||LEAVESBELOW[d][f]>1) {
					if(PLOTIUPAC)
						plotPCTNode(g, xOfNode[d][f], y+(PCTDEPTH-d+1)*LAYERDISTANCE+(PCTDEPTH-d)*NODEHEIGHT ,getIUPAC(NODELABELS[d][f]));
					else
						plotPCTNode(g, xOfNode[d][f], y+(PCTDEPTH-d+1)*LAYERDISTANCE+(PCTDEPTH-d)*NODEHEIGHT ,NODELABELS[d][f]);
				}
			}
		}
		plotPCTRootNode(g, x+NUMBEROFLEAVES*STACKWIDTH/2, y+LAYERDISTANCE);
	}
	
	private static void plotPCTNode(Graphics2D g, int x, int y, String label) {
		g.setColor( Color.BLACK );
		String text = "";
		for(int i=0;i<label.length();i++) {
			text+=label.charAt(i);
			if(i<(label.length()-1))
				text+=",";
		}
	//	int width = 12*label.length()+1*(label.length()-1)+12;
		int width = (int)(g.getFontMetrics().getStringBounds( text, g ).getWidth())+16;
		Font font = new Font(g.getFont().getName(),Font.PLAIN,NODELABELFONTSIZE);
		g.setFont( font );
		g.setColor( Color.WHITE );
		g.fillOval(x-width/2, y-NODEHEIGHT/2, width, NODEHEIGHT);
		g.setColor( Color.BLACK );
		g.drawOval(x-width/2, y-NODEHEIGHT/2, width, NODEHEIGHT);
	//	g.drawString(text, x-OFFSET[label.length()-1], y+NODEHEIGHT/4);
		g.drawString(text, x-(int)(g.getFontMetrics().getStringBounds( text, g ).getWidth()/2), y+NODEHEIGHT/4);
	}
	
	private static void plotPCTRootNode(Graphics2D g, int x, int y) {
		g.setColor( Color.BLACK );
		int width = NODEHEIGHT;
		Font font = new Font(g.getFont().getName(),Font.PLAIN,NODELABELFONTSIZE);
		g.setFont( font );
		g.setColor( Color.WHITE );
		g.fillOval(x-width/2, y-NODEHEIGHT/2, width, NODEHEIGHT);
		g.setColor( Color.BLACK );
		g.drawOval(x-width/2, y-NODEHEIGHT/2, width, NODEHEIGHT);
	}
	
	private static String getRoman(int number) {
       String res = "";  
       int rest = number;        
       for (int i = 0; i < numbers.length; i++) {
    	   while (rest >= numbers[i]) {
    		   res += letters[i];
               rest -= numbers[i];
               }
    	   }
       return res;
	}
	
	private static String getIUPAC(String label) {
		switch(label) {
			case "ACGT": return "N"; 
			case "ACG": return "V";
			case "ACT": return "T";
			case "AGT": return "D";
			case "CGT": return "B";
			case "AC": return "M";
			case "GT": return "K";
			case "AT": return "W";
			case "CG": return "S";
			case "CT": return "Y";
			case "AG": return "R";	
		}
		return label;
	}
}
