/*
    InMoDe is a collection of seven tools for learning, leveraging, and visualizing intra-motif dependencies within DNA binding sites and similar functional nucleotide sequences. 
    More information can be found at <http://jstacs.de/index.php/InMoDe>
    Copyright (C) 2017  Ralf Eggeling, Ivo Grosse, and Jan Grau
    
    InMoDe is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    InMoDe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, see <http://www.gnu.org/licenses>.
*/
package de.jstacs.projects.inmode.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Random;

import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.EmptyDataSetException;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.WrongLengthException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.data.alphabets.DiscreteAlphabet;
import de.jstacs.data.alphabets.DoubleSymbolException;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.data.sequences.SparseSequence;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.utils.Normalisation;

/**
 * Methods for conveniently reading and parsing input data.
 * @author Ralf Eggeling
 *
 */
public class DataReader {
	
	/**
	 * Reads a {@link File} parses its content into a matrix of {@link String} values.  
	 * @param f The input file
	 * @return A matrix of {@link String} values.  
	 * @throws IOException When the input file does not correspond to a matrix of {@link String} values.  
	 */
	public static String[][] readStringMatrix(File f) throws IOException  {
		BufferedReader br = new BufferedReader(new FileReader(f));
		String[][] res;
		int N=0;
		while(br.ready()) {
			br.readLine();
			N++;
		}
		br.close();
		br = new BufferedReader(new FileReader(f));
		res = new String[N][];
		for(int i=0;i<res.length;i++) {
			String s = br.readLine();
			if(s.contains("\t"))
				res[i] = s.split("\t");
			else
				res[i] = s.split(" ");
		}
		return res;	
	}
	
	/**
	 * Reads a {@link File} parses its content into a matrix of {@link Double} values.  
	 * @param f The input file
	 * @return A matrix of {@link Double} values.  
	 * @throws IOException When the input file does not correspond to a matrix of {@link Double} values.  
	 */
	public static double[][] readDoubleMatrix(File f) throws IOException {
		String[][] s = readStringMatrix(f);
		double[][] res = new double[s.length][];
		for(int i=0;i<res.length;i++) {
			res[i] = new double[s[i].length];
			for(int j=0;j<res[i].length;j++) {
				res[i][j] = Double.parseDouble(s[i][j]);				
			}
		}
		return res;
	}
	
	/**
	 * Reads a {@link File} parses its content into an array of {@link Double} values.  
	 * @param f The input file
	 * @return An array of {@link Double} values.  
	 * @throws IOException When the input file does not correspond to an array of {@link Double} values.  
	 */
	public static double[] readDoubleArray(File f) throws IOException {
		String[][] s = readStringMatrix(f);
		int L = 0;
		for(int i=0;i<s.length;i++) {
			L+=s[i].length;
		}
		double[] res = new double[L];
		int k=0;
		for(int i=0;i<s.length;i++) {
			for(int j=0;j<s[i].length;j++) {
				res[k] = Double.parseDouble(s[i][j]);
				k++;
			}
		}
		return res;
	}
	
	/**
	 * Reads a {@link File} parses its content into a matrix of {@link Integer} values.  
	 * @param f The input file
	 * @return A matrix of {@link Integer} values.  
	 * @throws IOException When the input file does not correspond to a matrix of {@link Integer} values.  
	 */
	public static int[][] readIntegerMatrix(File f) throws IOException {
		String[][] s = readStringMatrix(f);
		int[][] res = new int[s.length][];
		for(int i=0;i<res.length;i++) {
			res[i] = new int[s[i].length];
			for(int j=0;j<res[i].length;j++) {
				res[i][j] = Integer.parseInt(s[i][j]);
			}
		}
		return res;
	}
	
	/**
	 * Method for replacing non-ACGT IUPAC-symbols in DNA sequences. Every time a non-IUPAC symbol is encountered, it is replaced by a random sample from the nucleotide distribution in the data set, restricted to those possibilities that are consistent with the corresponding IUPAC symbol. 
	 * @param se Extractor of the original data file
	 * @return A {@link DataSet} over DNA alphabet.
	 * @throws Exception If something went wrong
	 */
	public static DataSet replaceIUPACSymbols (SparseStringExtractor se) throws Exception{
		DataSet ds = new DataSet( new AlphabetContainer(new DiscreteAlphabet( true, "A", "C", "G", "T", "N", "W", "S", "M", "K", "R", "Y", "B", "D", "H", "V", "X" )),se  );
		double[] prob = new double[4];
		for(int i=0;i<ds.getNumberOfElements();i++) {
			for(int j=0;j<ds.getElementAt(i).getLength();j++) {
				int val = ds.getElementAt(i).discreteVal(j);
				if(val<4) {
					prob[val]++;
				}
					
			}
		}
		prob = new double[]{1,1,1,1};
		Normalisation.sumNormalisation(prob);
		Random rand = new Random();
		Sequence[] seq = new Sequence[ds.getNumberOfElements()];
		for(int i=0;i<ds.getNumberOfElements();i++) {
			StringBuffer sb = new StringBuffer();
			for(int j=0;j<ds.getElementAt(i).getLength();j++) {
				int val = ds.getElementAt(i).discreteVal(j);
				if(val<4) {
					sb.append(ds.getAlphabetContainer().getSymbol(0, val));
				} else {
					if(val==4 || val==15) {
						sb.append(ds.getAlphabetContainer().getSymbol(0, BasicSamplingAlgorithms.sampleDiscreteRandomVariable(prob, rand)));
					} else {
						double[] pr = null ;
						if(val==5) pr = new double[]{prob[0],0,0,prob[3]};
						if(val==6) pr = new double[]{0,prob[1],prob[2],0};
						if(val==7) pr = new double[]{prob[0],prob[1],0,0};
						if(val==8) pr = new double[]{0,0,prob[2],prob[3]};
						if(val==9) pr = new double[]{prob[0],0,prob[2],0};
						if(val==10) pr = new double[]{0,prob[1],0,prob[3]};
						if(val==11) pr = new double[]{0,prob[1],prob[2],prob[3]};
						if(val==12) pr = new double[]{prob[0],0,prob[2],prob[3]};
						if(val==13) pr = new double[]{prob[0],prob[1],0,prob[3]};
						if(val==14) pr = new double[]{prob[0],prob[1],prob[2],0};
						Normalisation.sumNormalisation(pr);
						sb.append(ds.getAlphabetContainer().getSymbol(0, BasicSamplingAlgorithms.sampleDiscreteRandomVariable(pr, rand)));
					}
				}
			}
			seq[i] = new SparseSequence(new AlphabetContainer(DNAAlphabet.SINGLETON),sb.toString());	
			
		}
		return new DataSet( ds.getAnnotation(),seq);
	}

}
